#define HAVE_GETOPT_LONG /* XXX: Autodetect. Autoconf. F33r. --pasky */

#include <assert.h>
#include <getopt.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include "client/client.h"
#include "client/hud.h"
#include "client/players.h"
#include "client/render.h"
#include "common/bytef.h"
#include "common/conf.h"
#include "common/conn.h"
#include "common/debug.h"
#include "common/globgame.h"
#include "common/lists.h"
#include "common/packet.h"
#include "common/types.h"
#include "common/sprite.h"
#include "common/state.h"
#include "common/udp.h"
#include "common/xmalloc.h"
#include "gameio/gameio.h"
#include "version.h"

#define	MAXTRIES	30

#define	GAME_INIT	1
#define GAME_CONN	2
#define	GAME_WAIT	4
#define GAME_PREPARE	8
#define GAME_RUN	16

/**********/


int     team = 0;
char	*nickname = NULL;
char	talk_message[256];
int	talk_mode = 0;

int     paused = 0;

int     list_mode = 0;

enum game_style gstyle = GSTYLE_DM;
int     lms_lives = 0;


struct uconn *server;
struct conn	*chatserver;
u8	doexit = 0, cansend=0;
struct	timeval tv;

/* Options */
static int multispeed = 0;
static int kbd_driver_id, scr_driver_id;

/* Keybindings */
enum key_action {
	ACT_NONE,
	ACT_FIRE,
	ACT_UP, ACT_URIGHT, ACT_RIGHT, ACT_DRIGHT, ACT_DOWN, ACT_DLEFT, ACT_LEFT, ACT_ULEFT,
	ACT_ACCEL, ACT_DECCEL,
	ACT_LIST_PLAYERS, ACT_PAUSE, ACT_TALK, ACT_EXIT,
	ACT_MAX
};

static enum key_action std_binds[256];
static enum key_action raw_binds[ACT_MAX];
const char *key_actnames[] = {
	"none",
	"fire",
	"up", "uright", "right", "dright", "down", "dleft", "left", "uleft",
	"accel", "deccel",
	"list-players", "pause", "talk", "exit",
};

/* Direction and speed; speed is in power of 2 + 1, 0 is speed 0. */
int     speed = 0, key_direction = 0;
int	firing;

u16	last_ack = 0, cur_ack = 0;
struct stq	state;
u32	tick = 0;

/**********/

void	setexit(int sig)
{
	doexit++;
}

void	game_update(void)
{
	DEBUG("game update in state");

	tick++;
	if (double_kill > 0) double_kill--;
	if (multi_kill > 0) multi_kill--;

	tv.tv_sec = 0;
	tv.tv_usec = TICKMS*1000;
	udppck_send_cupdate(server, cur_ack, (speed<<3) + (firing?128:0) + key_direction);
	if (firing)
		firing--;
}

void	game_packet(struct udp_packet *pck)
{
	struct viewpoint *lvp, *cvp;
	switch(pck->type) {
		case HS_UPDATE:
			DEBUG("got c:%d l:%d vs c:%d l:%d", pck->su.curseq, pck->su.diffseq, cur_ack, last_ack);
			if (pck->su.curseq <= cur_ack) {
				DEBUG("ERROR tohle uz mam");
				return;
			}
			if (pck->su.diffseq > cur_ack) {
				DEBUG("ERROR tak tohle jeste nemam");
				return;
			}

			last_ack = pck->su.diffseq;
			cur_ack = pck->su.curseq;

			
			DEBUG("cur_ack: %d", cur_ack);
			rmlt_q(&state, last_ack);
			lvp = loc_q(&state, last_ack);
			cvp = addseq_q(&state, cur_ack);
			if (! lvp) {
				DEBUG("ERROR nemuzu najit viewpoint: %d", last_ack);
				dump_q(&state);
				return;
			}
 			if (!cvp) {
				DEBUG("ERROR nemuzu udelat viewpoint");
				dump_q(&state);
				return;
			}
			if (patch_view(lvp, cvp, pck->su.data, pck->su.size)) {
				DEBUG("ERROR cannot patch viewpoint");
				return;
			}
			render(cvp);
			return;
	}
	DEBUG("uknown packet opcode %x", pck->type);
}

void	sendmssg(u8 *msg, u16 len)
{
	u8 buffer[MAX_MSGLEN+4];

	if (len > MAX_MSGLEN-1)
		len = MAX_MSGLEN-1;
	DEBUG("::: SAY: '%s'", msg);
	bprintf(buffer, MAX_MSGLEN+4, "wq", CH_SAY, len, msg);
	write(chatserver->fd, buffer, len+4);
}


void    load_kbdbind(void)
{
	FILE *f;
	char line[128];
	int lineno;

	/* TODO: sysconfpath or smt. --pasky */
	f = fopen("kbdbind.conf", "r");
	if (!f) {
		perror("kbdbind.conf");
		exit(EX_UNAVAILABLE);
	}

	for (lineno = 1; fgets(line, 128, f); lineno++) {
		/* <char><sep><..action..> (<sep> can be anything, usually \t) */
		/* /^##/ is a comment; blank lines permitted */
		char	action[64];
		char	std_key, raw_key;
		char	raw_key_name[64];
		enum key_action i;

		raw_key_name[0] = 0;
		if (sscanf(line, "%[^ \t]%*c%c %s", action, &std_key, raw_key_name) < 2)
			continue;
//		printf("loaded bind for '%s', std: '%c' raw: '%s'\n", action, std_key, raw_key_name);
//		getchar();

		/* comment */
		if (action[0] == '#' && action[1] == '#')
			continue;

		if (raw_key_name[0])
			raw_key = gio_kbd_name2raw(raw_key_name);
		if (!raw_key) {
			fprintf(stderr, "kbdbind.conf: Unknown raw key name '%s' on line %d\n",
				raw_key_name, lineno);
			getchar();
			continue;
		}

		for (i = 0; i < ACT_MAX; i++) {
			if (!strcasecmp(action, key_actnames[i])) {
				std_binds[(int)std_key] = i;
				raw_binds[i] = raw_key;
				goto next_line;
			}
		}

		fprintf(stderr, "kbdbind.conf: Unknown action %s on line %d\n", action, lineno);
		getchar();
next_line:;
	}
}

void	handle_std_bind(int bind, char chr)
{
	if (talk_mode && chr > 0) {
		editor_char(chr);
		return;
	}

	switch (bind) {
		case ACT_FIRE:    firing = BULLET_TICKS; break;

		case ACT_UP:      key_direction = 0; break;
		case ACT_URIGHT:  key_direction = 1; break;
		case ACT_RIGHT:   key_direction = 2; break;
		case ACT_DRIGHT:  key_direction = 3; break;
		case ACT_DOWN:    key_direction = 4; break;
		case ACT_DLEFT:   key_direction = 5; break;
		case ACT_LEFT:    key_direction = 6; break;
		case ACT_ULEFT:   key_direction = 7; break;

		case ACT_ACCEL:
			if (multispeed) {
				speed++;
				if (speed > TANK_PSPEED + 1) speed = TANK_PSPEED + 1;
			} else {
				speed = TANK_PSPEED + 1;
			}
//			msg_add("Acceleration, tick %d", tick);
			break;

		case ACT_DECCEL:
			if (multispeed) {
				speed--;
				if (speed < 0) speed = 0;
			} else {
				speed = 0;
			}
			break;

		case ACT_LIST_PLAYERS:
			list_mode ^= 1;
			break;

		case ACT_PAUSE:
		{
			u8 buffer[4];

			DEBUG("::: PAUSE");
			bprintf(buffer, 2, "w", CH_PAUSE);
			write(chatserver->fd, buffer, 2);
		}
			break;

		case ACT_TALK:    talk_mode = 1; break;
		case ACT_EXIT:    doexit = 1; break;

		case ACT_NONE:    break;
		default:          assert(0); break;
	}
}
#define PRESSED(name) (rawkeys[raw_binds[(ACT_##name)]] != 0)
/* funny to use this */
#define RELEASED(name) \
{  \
	static int gonna_##name; \
	if (PRESSED(name)) gonna_##name = 1; \
		else if (gonna_##name) { \
			gonna_##name = 0;
 
void	handle_raw_binds(int key)
{
	int dir;
	static int trans[9] = { 7, 0, 1, 6, -1, 2, 5, 4, 3 };
	char chr = key & 0xff;
	
	/* firing */
	firing = PRESSED(FIRE);

	/* direction & movement */
	dir = trans[3*(PRESSED(DOWN)-PRESSED(UP)+1)+PRESSED(RIGHT)-PRESSED(LEFT)+1];
	if (dir != -1) {
		speed = TANK_PSPEED+1;
		key_direction = dir;
	} else {
		speed = 0;
	}
	if (talk_mode) {
		if ((chr) && (key > 0))
			editor_char(chr);
		return;
		/* we won't let the other stuff to take place as these are usually bound
		   to some letter key */
	}
	list_mode = PRESSED(LIST_PLAYERS);
	RELEASED(PAUSE) handle_std_bind(ACT_PAUSE, 0); }}
	RELEASED(TALK) talk_mode = 1; }}
	doexit = PRESSED(EXIT);	
}
void    kbd_poll()
{
	int	key;
	key = gio_kbd_key(tick);

	if (!kbd_is_raw) { /* keyboard for loosers */
		handle_std_bind(std_binds[key], key);
	} else
		handle_raw_binds(key);
	if ((!talk_mode) && (talk_message[0])) {
		sendmssg(talk_message, strlen(talk_message)+1);
		talk_message[0] = 0;
	}
	DEBUG("keys: %d %d %d", firing, speed, key_direction);
}


void	parse(void)
{
	u8	*ptr;
	u16	datalen, opcode, stringlen;
	int     scanlen;
	u16     id, teamid, victimid;
	s16     score, deaths, fwon, flost, fstolen;
	char    string[256];

	/* check na bounds ... */
	while (1) {
		struct buffer *pktbuf = chatserver->in;

		ptr = pktbuf->data;
		datalen = BUFFER_NDATA(chatserver->in);
		if (datalen < 2) {
			DEBUG("Too short packet %d<%d", datalen, 2);
			return;
		}

		opcode = UNPACKW(ptr);
		ptr += 2; datalen -= 2;

		/* TODO: More strict rangechecking of input data (especially
		 * ids). --pasky */

		DEBUG("[parse] data: %d, opcode: %x", datalen, opcode);
		switch (opcode) {
			case CH_MSGGSTYLE:
			{
				static const char *stylenames[GSTYLE_MAX] = {
					"deathmatch",
					"last man standing",
					"capture the flag",
				};
				int totlen = 0;

				/* This is actually pretty ugly. It might be
				 * worthwhile to just have a special opcode for
				 * each game style. --pasky */

				scanlen = bscanf(ptr, datalen, "w", &gstyle);
				if (scanlen < 0) {
					DEBUG("Too short gstyle packet %d", datalen);
					return;
				}
				ptr += scanlen; datalen -= scanlen; totlen += scanlen;

				if (gstyle >= GSTYLE_MAX) {
					DEBUG("Invalid game style %d!", gstyle);
					return;
				}
				msg_add("The game style is %s.", stylenames[gstyle]);
				DEBUG(":::: MGSTYLE: gstyle%d", gstyle);

				switch (gstyle) {
					case GSTYLE_DM:
					       	break;
					case GSTYLE_LMS:
						scanlen = bscanf(ptr, datalen, "w", &lms_lives);
						if (scanlen < 0) {
							DEBUG("Too short LMS packet %d", datalen);
							return;
						}
						ptr += scanlen; datalen -= scanlen; totlen += scanlen;
						msg_add("You have %d lives.", lms_lives);
						break;
					case GSTYLE_CTF:
						break;
					default: assert(0); break;
				}
				buffer_discard(pktbuf, 2 + totlen);
			}
				break;

			case CH_MSGID:
				BSCANF("msgid", "w", &id);

				my_id = id;
				if (!self_player) self_player = find_player(id);
				DEBUG(":::: MSGID: id%d", id);
				udppck_send_hello(server, id);
				break;

			case CH_MSGJOIN:
				stringlen = MAX_NICKLEN;
				BSCANF("join", "wwwq", &id, &score, &deaths, &stringlen, &string);

				msg_add("%s has joined the game.", string);
				DEBUG(":::: MJOIN: id%d score%d deaths%d '%s'", id, score, deaths, string);
				{
					struct player_info *plr = add_player(id, string, score, deaths);
					if (!self_player && my_id == id)
						self_player = plr;
				}
				break;

			case CH_MSGCHAT:
				stringlen = MAX_MSGLEN;
				BSCANF("chat", "wq", &id, &stringlen, &string);

			{
				struct player_info *plr = find_player(id);

				msg_add("<%c%s%c> %s", PRINTF_COLOR(plr_color(plr->team_id ? (plr->team_id+7)%8 : plr->id)),
				        plr->nick, PRINTF_COLOR(0), string);
				DEBUG(":::: MCHAT id%d '%s'", id, string);
			}
				break;

			case CH_MSGKILL:
				BSCANF("kill", "ww", &victimid, &id);

				{
					struct player_info *plr = find_player(id);
					struct player_info *victim = find_player(victimid);

					if (victim != plr) {
						msg_add("%s got fragged by %s.",
						        victim == self_player ? "You" : victim->nick,
							plr->nick);
					} else {
						msg_add("%s got shredded by own missile.",
						        victim == self_player ? "You" : plr->nick);
					}
					kill_player(victim, plr);
				}
				DEBUG(":::: MKILL id%d by id%d", victimid, id);
				break;

			case CH_MSGTELEFRAG:
				BSCANF("telefrag", "ww", &victimid, &id);

				if (victimid == id) {
					DEBUG("Self-telefrag [%d] is impossible.", id);
					/* At least until we have some real teleportation. */
					return;
				}
				{
					struct player_info *plr = find_player(id);
					struct player_info *victim = find_player(victimid);

					/* Keep in sync with server code! */
					if (victim == self_player)
						msg_add("Suddenly a solid metal block phases into your cockpit!");
					msg_add("%s got telefragged by %s.",
						victim == self_player ? "You" : victim->nick,
						plr->nick);
					kill_player(victim, plr);
				}
				DEBUG(":::: MTELEFRAG id%d by id%d", victimid, id);
				break;

			case CH_MSGEDEATH:
				BSCANF("edeath", "w", &id);

				{
					struct player_info *plr = find_player(id);

					if (plr == self_player)
						msg_add("Eventually a thundering explosion rips through the whole tank.");
					msg_add("%s%s overburdened energy generator exploded.",
						plr == self_player ? "You" : plr->nick,
						plr == self_player ? "r" : "'s");
					kill_player(plr, plr);
				}
				DEBUG(":::: MEDEATH id%d", id);
				break;

			case CH_MSGGLAMOR:
				BSCANF("glamor", "ww", &id, &score);

				{
					struct player_info *plr = find_player(id);
					static const char *up[] = {
						"%s is on a killing spree!",
						"%s is awarded by the supreme command!",
						"%s rules the battlefield!",
						"%s seems to be deathless!",
					};
					static const char *down[] = {
						"%s's killing spree is over.",
						"The supreme command loses interest in %s.",
						"%s no more rules the battlefield.",
						"The death indeed have found %s at last.",
					};
					static const char *banners[] = {
						"K I L L I N G   S P R E E",
						"A W A R D E D",
						"T H E   R U L E R",
						"D E A T H L E S S",
					};

					DEBUG("glamor %d -> %d", plr->glamor, score);
					if (score > 0) {
						assert(score < 5);
						msg_add(up[score - 1], plr->nick);
						if (plr == self_player)
							banner = banners[score - 1];
					} else if (plr->glamor > 0) {
						msg_add(down[plr->glamor - 1], plr->nick);
						if (plr == self_player)
							banner = NULL;
					}
					plr->glamor = score;
				}
				DEBUG(":::: MGLAMOR id%d level%d", id, score);
				break;

			case CH_MSGPAUSE:
				BSCANF("pause", "ww", &id, &score);

				msg_add(score ? "%s got partially phased out from the battle plane."
				              : "%s got phased back to the battle plane.",
				        self_player && id == self_player->id ? "You" : player_nick_by_id(id));
				if (self_player && id == self_player->id)
					paused = score;
				break;

			case CH_MSGWINNER:
				BSCANF("winner", "w", &id);

				{
					struct player_info *plr = find_player(id);

					if (plr == self_player)
						msg_add("The whole battlefield bows to... you!");
					msg_add("%s %s THE GAME WINNER!",
						plr == self_player ? "YOU" : plr->nick,
						plr == self_player ? "ARE" : "IS");
					foreach (plr, players) {
						/* Won't break sorting. */
						plr->score = 0;
						plr->deaths = 0;
						plr->glamor = 0;
					}
				}
				DEBUG(":::: MWINNER id%d", id);
				break;

			case CH_MSGLEAVE:
				BSCANF("leave", "w", &id);

				msg_add("%s left the game.", player_nick_by_id(id));
				DEBUG(":::: MLEAVE id%d", id);
				del_player(find_player(id));
				break;

			case CH_MSGTEAMUP:
				BSCANF("teamup", "ww", &id, &teamid);

				if (teamid >= MAX_TEAMS) {
					DEBUG("ERROR - invalid team ID %d", teamid);
					return;
				}
			{
				struct player_info *plr = find_player(id);

				DEBUG(":::: MTEAMUP id%d team%d", id, teamid);
				plr->team_id = teamid;
				if (!teamid) {
					msg_add("%s is playing freelance now.", plr->nick);
				} else {
					msg_add("%s joined %s.", plr->nick, teams[teamid].name);
					if (!teams[teamid].active) {
						teams_count++;
						teams[teamid].active++;
					}
				}
			}
				break;

			case CH_MSGTEAMINFO:
				stringlen = MAX_NICKLEN;
				BSCANF("teamup", "wwwqwww", &teamid, &score, &deaths,
				       &stringlen, &string, &fwon, &flost, &fstolen);

				if (teamid >= MAX_TEAMS) {
					DEBUG("ERROR - invalid team ID %d", teamid);
					return;
				}
				teams[teamid].score = score;
				teams[teamid].deaths = deaths;
				teams[teamid].fwon = fwon;
				teams[teamid].flost = flost;
				teams[teamid].fstolen = fstolen;
				if (teams[teamid].active) {
					msg_add("Team %s was renamed to %s.", teams[teamid].name, string);
				} else {
					teams_count++;
				}
				if (teams[teamid].name) free(teams[teamid].name);
				teams[teamid].name = strdup(string);
				teams[teamid].active++;

				DEBUG(":::: MTEAMINFO teamid%d score%d deaths%d fwon%d flost%d fstl%d name%s",
				      teamid, teams[teamid].score, teams[teamid].deaths,
				      teams[teamid].fwon, teams[teamid].flost, teams[teamid].fstolen,
				      teams[teamid].name);
				break;

			case CH_MSGFLAGPICK:
				BSCANF("flagpick", "ww", &id, &teamid);

				if (teamid >= MAX_TEAMS) {
					DEBUG("ERROR - invalid team ID %d", teamid);
					return;
				}
			{
				struct player_info *plr = find_player(id);

				DEBUG(":::: MFLAGPICK id%d team%d", id, teamid);
				plr->flagged = 1;
				teams[teamid].fstolen = 1;
				msg_add("%s grabbed up the flag of team %s.", plr->nick, teams[teamid].name);
			}
				break;

			case CH_MSGFLAGDROP:
				BSCANF("flagdrop", "ww", &id, &teamid);

				if (teamid >= MAX_TEAMS) {
					DEBUG("ERROR - invalid team ID %d", teamid);
					return;
				}
			{
				struct player_info *plr = find_player(id);

				DEBUG(":::: MFLAGDROP id%d team%d", id, teamid);
				plr->flagged = 0;
				msg_add("%s dropped the flag of team %s.", plr->nick, teams[teamid].name);
			}
				break;

			case CH_MSGFLAGRET:
				BSCANF("flagret", "ww", &id, &teamid);

				if (teamid >= MAX_TEAMS) {
					DEBUG("ERROR - invalid team ID %d", teamid);
					return;
				}
			{
				struct player_info *plr = find_player(id);

				DEBUG(":::: MFLAGRET id%d team%d", id, teamid);
				teams[teamid].fstolen = 0;
				msg_add("%s returned the flag of team %s to the base.", plr->nick, teams[teamid].name);
			}
				break;

			case CH_MSGFLAGWIN:
				BSCANF("flagwin", "ww", &id, &teamid);

				if (teamid >= MAX_TEAMS) {
					DEBUG("ERROR - invalid team ID %d", teamid);
					return;
				}
			{
				struct player_info *plr = find_player(id);

				DEBUG(":::: MFLAGWIN id%d team%d", id, teamid);
				plr->flagged = 0;
				teams[plr->team_id].fwon++;
				teams[teamid].fstolen = 0;
				teams[teamid].flost++;
				msg_add("%s fetched the flag of team %s to his base and scores!",
				        plr->nick, teams[teamid].name);
			}
				break;

			default:
				return;
		}
	}
}



/**********/

void	mainloop(void)
{
	fd_set	fdr, fdt;
	s32	ret;
	struct udp_packet	packet;

	FD_ZERO(&fdr);
	FD_SET(server->fd, &fdr);
	FD_SET(chatserver->fd, &fdr);

	tv.tv_sec = tv.tv_usec = 0;
	init_teams();
	if (gio_kbd_init(kbd_driver_id, 0, 1) < 0)
		exit(1);
	if (gio_scr_init(scr_driver_id, VIEWPNT_WIDTH, VIEWPNT_HEIGHT) < 0) {
		gio_kbd_cleanup();
		exit(1);
	}
	render_init(VIEWPNT_WIDTH, VIEWPNT_HEIGHT);

	init_q(&state, 10000, 10000);

	while (! doexit) {
		fdt = fdr;
		ret = select(FD_SETSIZE, &fdt, NULL, NULL, &tv);
		kbd_poll();
		DEBUG("select done, ret:%d sec: %ld usec: %ld", ret, tv.tv_sec, tv.tv_usec);
		if (ret < 0)
			continue;
		else if (ret == 0) {
			if (cansend)
				game_update();
			else
				tv.tv_sec = 1;
		} else if (FD_ISSET(server->fd, &fdt)) {
			cansend = 1;
			DEBUG("*** UDP PACKET");
			while (udppck_recv(server, &packet))
				game_packet(&packet);
		} else if (FD_ISSET(chatserver->fd, &fdt)) {
			int l;

			DEBUG("*** TCP ACTIVITY");
			if ((l = buffer_read(chatserver->in, chatserver->fd)) <= 0) {
				DEBUG("`--- QUITING (%d) (%s)", l, (l < 0 ? strerror(errno) : "no data"));
				doexit = 1;
			} else  {
				DEBUG("`--- PARSING data");
				parse();
			}
		}
	}

	/* Try to do the final render. Useful when the server closed connection
	 * but said why right before. */
	if (last_ack) render(loc_q(&state, last_ack));

	gio_scr_cleanup();
	gio_kbd_cleanup();
	udp_close(server);
	conn_destroy(chatserver);
	printf("bye...\n");
	exit(EX_OK);
}



static void
load_client_config(int *uport, int *fps)
{
	load_cfg_vars("client.cfg", "nick fps udp-port multispeed team",
	                     "s i i i i",
	                     &nickname, fps, uport, &multispeed, &team);
}

char version_banner[256];

static void
usage(char *argv[])
{
#define list_drivers(type_) \
	{ \
		struct gi_##type_##_driver **drv; \
		for (drv = type_##_drivers; *drv; drv++) { \
			printf("                             %-5s   %s%s.\n", \
			       (*drv)->name, (*drv)->desc, drv == type_##_drivers ? " (default)" : ""); \
		} \
	}
	printf("%s\n", version_banner);
	printf("Usage: %s [OPTION]... SERVER_ADDR [TCPPORT]\n", argv[0]);
	printf("\n");
	printf("Mandatory arguments to long options are mandatory for short options too.\n");
	printf("\n");
	printf("Default TCPPORT value is %d.\n", DEFAULT_PORT);
	printf("\n");
	printf(" -dk, --kbd-driver=NAME    use specified keyboard driver:\n");
	list_drivers(kbd);
	printf(" -ds, --scr-driver=NAME    use specified screen driver:\n");
	list_drivers(scr);
	printf("  -f, --fps=NUMBER         how many FPS to show in the client.\n");
	printf("  -n, --nick=NAME          your nickname to use.\n");
	printf("  -P, --udp-port=UDPPORT   bind specified UDPPORT; by default\n");
	printf("                           the same port as TCPPORT is used.\n");
	printf("  -s, --multispeed         speed is increased/decreased in steps.\n");
	printf("  -t, --team=NUMBER        join the given team (max: %d).\n", MAX_TEAMS - 1);
	printf("  -h, --help               show usage information.\n");
	printf("\n");
	printf("Please report bugs and suggestions on #tunneler @ IRCNet.\n");
	exit(EX_USAGE);
#undef list_drivers
}

int	main(int argc, char *argv[])
{
	char *server_addr = NULL;
	int tport = -1, uport = -1, fps = FPS;
	int show_usage = 0, unnamed_pos = 0, expecting_driver = 0;
	u8 buffer[128], *bufp;

	snprintf(version_banner, 256, "Tunneler client%s version udp-%04d -- in 2005 by ment, pasky, sd",
	         version_debug, version);

	load_client_config(&uport, &fps);

	while (1) {
#ifdef HAVE_GETOPT_LONG
		static struct option long_options[] = {
			{ "kbd-driver", 1, 0, 2   },
			{ "scr-driver", 1, 0, 3   },
			{ "fps",        1, 0, 'f' },
			{ "help",       0, 0, 'h' },
			{ "nick",       1, 0, 'n' },
			{ "udp-port",   1, 0, 'P' },
			{ "multispeed", 0, 0, 's' },
			{ "team",       1, 0, 't' },
			{ NULL,         0, 0, 0 }
		};
		int optindex;
#endif
		char *numguard = NULL;
		int c;
#define invalid_option(name_) \
		do { \
			fprintf(stderr, "%s: invalid "name_" -- %s\n", argv[0], optarg); \
			show_usage = 1; \
		} while (0)
#define load_number(name_, dest_) \
		do { \
			dest_ = strtol(optarg, &numguard, 10); \
			if (*numguard) invalid_option(name_); \
		} while (0)


#ifdef HAVE_GETOPT_LONG
		c = getopt_long(argc, argv, "-d:f:hn:P:st:", long_options, &optindex);
#else
		c = getopt(argc, argv, "-d:f:hn:P:st:");
#endif
		if (c == -1) break;

		switch (c) {
			case 1:
				if (expecting_driver) {
					switch (expecting_driver) {
					case 'k':
set_kbd_driver:
						kbd_driver_id = gio_kbd_find_driver(optarg);
						if (kbd_driver_id < 0)
							invalid_option("NAME");
						break;
					case 's':
set_scr_driver:
						scr_driver_id = gio_scr_find_driver(optarg);
						if (scr_driver_id < 0)
							invalid_option("NAME");
						break;
					}
					expecting_driver = 0;
					break;
				}
				switch (unnamed_pos++) {
					case 0: server_addr = optarg; break;
					case 1: load_number("TCPPORT", tport); break;
					default: show_usage = 1; break;
				}
				break;
			case 2: goto set_kbd_driver;
			case 3: goto set_scr_driver;
			case 'd':
				/* Trick idea by wget. */
				if (strlen(optarg) != 1) {
					invalid_option("-d option");
					break;
				}
				expecting_driver = *optarg;
				break;
			case 'f': load_number("NUMBER", fps); break;
			case 'n': nickname = optarg; break;
			case 'P': load_number("UDPPORT", uport); break;
			case 's': multispeed = 1; break;
			case 't': load_number("NUMBER", team); break;
			case 'h':
			case '?': show_usage = 1; break;
		}
#undef load_number
#undef invalid_option
	}

	if (!nickname) {
		fprintf(stderr, "%s: nickname required.\n", argv[0]);
		usage(argv);
	}
	if (!server_addr || show_usage) usage(argv);
	if (tport < 0) tport = DEFAULT_PORT;
	if (uport < 0) uport = tport;


	debuglog("client.log");
	DEBUG("%s starting...", version_banner);
	load_kbdbind();
	init_list(players);

	load_default_gfx();

	srandom(time(NULL));


	server = udp_client(server_addr, uport);
	if (server->state == UCS_ERROR) {
		perror("socket");
		exit(EX_OSERR);
	}

	chatserver = conn_connect_resolv(server_addr, tport);
	if (chatserver->state != CS_CONNECTED) {
		perror("connect");
		exit(42);
	}
	conn_setbuff(chatserver, 900, 0);

	signal(SIGTERM, setexit);
	signal(SIGKILL, setexit);
	signal(SIGINT, setexit);
	signal(SIGPIPE, SIG_IGN);


	DEBUG("connected, sending hello");
	bufp = bprintf(buffer, 128, "wwws", CH_JOIN, fps, team, nickname);
	if (!bufp) {
		printf("cannot format buffer!?\n");
		exit(EX_SOFTWARE);
	}
	write(chatserver->fd, buffer, bufp-buffer);

	DEBUG("socket created, looping");
	mainloop();
	return EX_OK;
}
