#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "client/client.h"
#include "client/hud.h"
#include "client/players.h"
#include "client/render.h"
#include "common/debug.h"
#include "common/globgame.h"
#include "common/math.h"
#include "common/viewpoint.h"
#include "gameio/colors.h"

/* This code takes care of the basic HUD rendering. */


/* color of bars */
#define CRIT_COLOR (MAKEBG(COLOR_RED)|COLOR_DARKGRAY)
#define SH_COLOR (MAKEBG(COLOR_GREEN)|COLOR_DARKGRAY)
#define EN_COLOR (MAKEBG(COLOR_BLUE)|COLOR_DARKGRAY)


const char *banner = NULL;

/* Prints colored banner at the bottom; pass some tick counter (i.e. @ticks)
 * as blink if you wanna it to blink, 0 otherwise. */
static void
banner_print(int color, long blink, const char *text)
{
	if ((blink * 4 / FPS) % 2) {
		if (color > 7)
			color -= 8;
		else
			return;
	}

	screen_printf(((VIEWPNT_WIDTH-4-2-10-2-39 - strlen(text)) / 2)-2+39,
	              VIEWPNT_HEIGHT-1, color, "%s", text);
}


static void
render_bars(struct viewpoint *vp)
{
	int meterw = VIEWPNT_WIDTH-18;
	screen_darken(2+12, VIEWPNT_HEIGHT-3, (meterw * vp->en) / 255, (vp->en<64)?CRIT_COLOR:EN_COLOR);
	screen_printf(4, VIEWPNT_HEIGHT-3, COLOR_WHITE, "Energy  : %c%03d%%",
			PRINTF_COLOR(COLOR_YELLOW), vp->en * 100 / 255);
	screen_darken(2+12, VIEWPNT_HEIGHT-2, (meterw * vp->sh) / 255, (vp->sh<64)?CRIT_COLOR:SH_COLOR);
	screen_printf(4, VIEWPNT_HEIGHT-2, COLOR_WHITE, "Shields : %c%03d%%",
			PRINTF_COLOR(vp->sh<64?COLOR_YELLOW:COLOR_RED), vp->sh * 100 / 255);
}

static void
render_status(struct viewpoint *vp)
{
	screen_printf(4, VIEWPNT_HEIGHT-1, COLOR_WHITE, "Frags : %2d", self_player ? self_player->score : 0);
	if (gstyle == GSTYLE_LMS) {
		screen_printf(20,VIEWPNT_HEIGHT-1, COLOR_WHITE, "Lives : %2d", self_player ? lms_lives - self_player->deaths : 0);
	} else {
		screen_printf(20,VIEWPNT_HEIGHT-1, COLOR_WHITE, "Deaths : %2d", self_player ? self_player->deaths : 0);
	}
	screen_printf(VIEWPNT_WIDTH-4-1-10, VIEWPNT_HEIGHT-1, COLOR_WHITE, "Speed : %3d", (speed == 0 ? 0 : 1 << (speed - 1)) * 5);
}


static void
render_quip(struct viewpoint *vp)
{
	static int ad_wait = -1, ad_show = -1, quip_show = -1;
#define QUIPS_COUNT 149
	static const char *quips[QUIPS_COUNT] = {
		"Surprise 'em.", "Report?", "20 from the north!",
		"Report, commander.", "You aren't on my radar!",
		"We can't hear you!", "Watch more TV.", "Surrender!",
		"m3 4 l33t h4x0r?", "3 from the east!", "Hurry up...",
		"You're not alone.", "News at 11, live.", "It's over.",
		"Save the whales!", "We have a problem.", "Report!",
		"The monkey did it.", "He has a nuke!", "Oh god...",
		"You're in trouble.", "Feeling lonely?", "Watch this!",
		"To. m.ch nois....", "^%4 @.!1 *(## 9a+~ 4]z",
		"Fantomas was here.", "Abg sbe lbh.", "Better go.",
		"Out of ammunition.", "7 from the south!", "Pink?! ;-)",
		"You must be pretty good!", "D1s i2 r33ly l33t!",
		"Not another time...", "The end is coming!", "EAGAIN",
		"The truth is out there.", "I can do better.",
		"That's wild!", "Use ELinks, too.", "Marvelous!",
		"Help! Over here!", "Plleeassee... *kaboom*", "brm",
		"I didn't do that.", "Me, the HUD gnome!", "*TCHUNK*",
		"ATDT+420723300133", "Freedom for the animals!",
		">clickety click<", "Segmentation fault", "SIGFPU",
		"You are gonna *DIE*!", "That was stupid.", "LoL",
		"Watch out.", "I like this game too.", "Bah...",
		"...e can't h...r y..u...", "Excellent!", "Stop it.",
		"Special level!", "That chick was hot!", "CHEATS ON",
		"That's pretty dangerous.", "I don't really know.",
		"I wouldn't do that.", "Wrong answer!", "Think so?",
		"*beep beep beep beep*", "He's better, anyway.",
		"You're already dead.", "You don't have a chance.",
		"The weather is nice.", "Out of ammunition.",
		"They ran that way.", "4 from the west!", "C2 to C4",
		"Enemy approaching!", "Ye'rt doomed!", "*burp*",
		"He will kick you ass.", "It's all red!",
		"Life is just a dream.", "We're tracking your pos.",
		"I saw him first!", "Something smells here.",
		"Isn't this smoke?", "Isn't it too hot here?",
		"Locked on your signal.", "Fire at will!",
		"Telefrag him.", "Kill 'em all!", "Let them die!",
		"Ouch.", "Oops.", "Sir?", "Sir!", "Got him!", "*",
		"Wo bin ich?", "Friendly fire!", "#", "lim(1+1/x)^x",
		"Interesting.", "Go away.", "Get lost.", "You suck.",
		"Ja nepanemaju!", "Nu da...", "Scheisse!",
		"Press any key.", "Aber das ist wenig!", "Press 'q'!",
		"Play more ADOM!", "Scorch? Scorch!", "*yawn*",
		"<pasky> IRC gw up", "The last is the best.",
		"ATDT+420272916775", "Did you see it too?",
		"Leave the ship!", "Lieutenant?", "That was me!",
		"Oh no, not again.", "Quote of the day: '",
		"SERVER IS OVERLOADED", "YOUR BOX IS HACKED",
		"Pick the time travel special.", "bash-2.05#",
		"I violate patent #5487022461.", "Don't listen to him!",
		"This is the USS Enterprise.", "Leave the vehicle!",
		"Leave the tank!", "You are in grave danger.", "f33r",
		"100% ASCII", "But the artwork is nice.", "Gruesome.",
		"EC31475 fx p0 r14 3z", "Ignoring SIGSEGV", "Seven.",
		"DEBUG: 0x47a509dc 1/3::0x4a", "All around you.",
		"This is a trap!", "..un.er ..eavy f..re!",
		"It's wet here.", "You just ran over a mole!",
		/* Feel free to add more. --pasky */
	};

	if (ad_show-- >= 0) {
		banner_print(COLOR_GREEN, 0, quips[quip_show]);

	} else if (ad_wait < 0) {
		ad_wait = 5 * FPS + mkrand(FPS * 120);

	} else if (ad_wait-- == 0) {
		ad_show = FPS / 2 + mkrand(FPS * 3);
		quip_show = mkrand(QUIPS_COUNT);
	}
#undef QUIPS_COUNT
}

static void
render_banner(struct viewpoint *vp)
{
	if (multi_kill) {
		banner_print(/* red */ COLOR_LIGHTRED, multi_kill, "M U L T I K I L L !");
	} else if (double_kill) {
		banner_print(/* red */ COLOR_LIGHTRED, double_kill, "D O U B L E K I L L !");
	} else if (gstyle == GSTYLE_CTF && self_player && self_player->flagged) {
		banner_print(/* red */ COLOR_LIGHTRED, tick, "F E T C H   T H E   F L A G !");
	} else if (banner) {
		banner_print(/* red */ COLOR_LIGHTRED, 0, banner);
	} else if (gstyle == GSTYLE_LMS && self_player && self_player->deaths == lms_lives - 1) {
		banner_print(/* red */ COLOR_LIGHTRED, 0, "L A S T   C H A N C E !");
	} else if (gstyle == GSTYLE_CTF && self_player && teams[self_player->team_id].fstolen) {
		banner_print(/* red */ COLOR_LIGHTRED, 0, "R E C O V E R   F L A G !");
	} else {
		render_quip(vp);
	}
}


void	render_hud(struct viewpoint *vp)
{
	screen_darken(2, VIEWPNT_HEIGHT-3, VIEWPNT_WIDTH-4, COLOR_DARKGRAY);
	screen_darken(2, VIEWPNT_HEIGHT-2, VIEWPNT_WIDTH-4, COLOR_DARKGRAY);
	screen_darken(2, VIEWPNT_HEIGHT-1, VIEWPNT_WIDTH-4, COLOR_DARKGRAY);

	render_bars(vp);
	render_status(vp);
	render_banner(vp);
}
