//#define ALT_DIRT
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "client/client.h"
#include "client/hud.h"
#include "client/players.h"
#include "client/render.h"
#include "common/debug.h"
#include "common/globgame.h"
#include "common/packet.h"
#include "common/sprite.h"
#include "common/types.h"
#include "common/viewpoint.h"
#include "gameio/gameio.h"

/* max no of msgs*/
#define MSGS_MAX	15 /* to accomodate initial burst */
#define MSGS_GOOD       5
#define MSGS_TIMEOUT	15

char	msgs[MSGS_MAX][VIEWPNT_WIDTH];
u32	msgt[MSGS_MAX];
int	msglast = 0;
int	msgshow = 0;

pixel *scr;
extern	u32 tick;

/* [8] is special, for paused tank. */
u8	colors[9] = { BRIGHT(COLOR_RED), BRIGHT(COLOR_GREEN),
                      BRIGHT(COLOR_BROWN), BRIGHT(COLOR_BLUE),
		      BRIGHT(COLOR_MAGENTA), BRIGHT(COLOR_CYAN),
		      BRIGHT(COLOR_GRAY), COLOR_BROWN, BRIGHT(COLOR_BLACK) };

int	plr_color(int id)
{
	return colors[id%8];
}

/* render one sprite */
void	render_sprite(struct gfx_object *sprite, s16 sx, s16 sy, int state, int color)
{
	struct sprite *s;
	u16 	i, j;

	s = sprite->directions[state]->frames[0];
	//DEBUG("draw x:%d y:%d state:%d h:%d", sx, sy, state, s->height);
	for (j=0; j<s->height; j++) {
		//DEBUG("j %d s %d V %d", j, sy, VIEWPNT_HEIGHT);
		for (i=0; i<s->width; i++) {
			struct sprite_char c = s->sprite[j*s->width+i];
			pixel orig = GET_PIXEL_CHECK(scr, (i+sx), (j+sy));
			char ptext = PIXCHR(orig), pcolor = PIXFG(orig);

			//DEBUG("%c/%d/%d", c.text, c.attr, c.color);
			if (!(sprite_char_transparent(c)))
				ptext = c.text;
			switch (c.attr) {
				case SCA_NORMAL:      pcolor = color; break;
				case SCA_STANDOUT:    pcolor = color^8; break;
				case SCA_TRANSPARENT: if (c.text != ' ') {
							ptext = c.text;
							pcolor = color;
						       }
							break;
				case SCA_TRANSLUCENT: if (pcolor == 7) pcolor = 8;
						      else if (pcolor != 8) pcolor &= ~8; break;
				case SCA_DARKCOLOR:   pcolor = c.color; break;
				case SCA_BRIGHTCOLOR: pcolor = c.color|8; break;
			}
			PUT_PIXEL_CHECK(scr, (i+sx), (j+sy), MAKEPIX(ptext, pcolor));
		}
	}
}

#define HASH_DIRT(x,y) ((unsigned char) ((x<<5) ^ (y>>2) + x - y))
#define HASH_DIRT_COLOR(x,y) (((unsigned char) (x^y))&1)
#define HASH_ROCK(x,y) ((unsigned char)((x^y)*2+x+y+tick_rock))
#define HASH_HOME(x,y) ((unsigned char)(x+y+tick_home))

/* precompute sin/cos values for plasma-effect dirt rendering */

#define PLASMA_X 512
#define PLASMA_Y 256
#define PLASMA_ROUNDS 7
#define RANGE	256
#define DENSITY 80
#define SIZE	200

/* how much levels will plasma have */
#define PLEVELS	8

int	plasmamap[PLASMA_Y][PLASMA_X];
int	randtab[4096];

static	inline int plasma_pix(int x, int y)
{
	if (x >= PLASMA_X)
		x = PLASMA_X-x;
	if (y >= PLASMA_Y)
		y = PLASMA_Y-y;
	if (x < 0)
		x = PLASMA_X+x;
	if (y < 0)
		y = PLASMA_Y+y;
	return plasmamap[y][x];
}

static	void	plasma_init()
{
	int x, y, i;
	for (i = 0; i < 4096; i++)
		randtab[i] = rand();
	/* create map */
	for (y = 0; y < PLASMA_Y; y++) {
		for (x = 0; x < PLASMA_X; x++) {
			if (!(rand()%DENSITY)) {
				plasmamap[y][x] = RANGE*SIZE + (rand() % (RANGE*SIZE));
			} else 
				plasmamap[y][x] = 0;
		}
	}
	/* and soften it */
	for (i = 0; i < PLASMA_ROUNDS; i++)
	for (y = 0; y < PLASMA_Y; y++) {
		for (x = 0; x < PLASMA_X; x++) {
			plasmamap[y][x] =
			(plasma_pix(x-1, y-1) + plasma_pix(x, y-1) + plasma_pix(x+1, y-1) +
			 plasma_pix(x-1, y) /* + plasma_pix(x, y) */ + plasma_pix(x+1, y) + 
			 plasma_pix(x-1, y+1)  + plasma_pix(x, y+1) + plasma_pix(x+1, y+1)) / 9;
		}
	}
	for (y = 0; y < PLASMA_Y; y++) {
		for (x = 0; x < PLASMA_X; x++) {
			plasmamap[y][x] = plasmamap[y][x]*PLEVELS/RANGE;
			if (plasmamap[y][x] >= PLEVELS)
				plasmamap[y][x] = PLEVELS-1;
		}
	}
}

#define PLASMA(x,y) (((plasmamap[y%PLASMA_Y][x%PLASMA_X]))%PLEVELS)
#define NDIRT	8

#define RND1 16333
#define RND2 25887
#define RND3 17

/* ok, let's try something more sophisticated - error distribution (whoa!) */
static	inline	pixel compute_plasma(int x, int y)
{
	unsigned hash = randtab[(y*5001+x)%4096];
	//((x+y)*RND1+(RND2)) % RND3;
	int	level = PLEVELS-PLASMA(x,y); /* 1 highest level */
	static	char	dirt[] = "...,,,:;``''";
	static	char    dirthi[] = ".,:;;\"++";

	if (level*6 > (PLEVELS*5))
		return MAKEPIX(dirt[hash%(sizeof(dirt)-1)], BRIGHT(COLOR_BLACK));
	return MAKEPIX(dirthi[hash%(sizeof(dirthi)-1)], (!(hash%level))?COLOR_GRAY:BRIGHT(COLOR_BLACK));
}

/* render dirt, rock etc */
void	render_view(u8 *screen, s16 xo, s16 yo)
{
	u8 	c;
	u16	x,tx, y,ty, k=0;
#if 0
	int	tick_rock = tick>>2;
	int	tick_home = tick>>1;
#endif
#ifdef ALT_DIRT
	int	hash;
#endif

	for (ty = yo, y=0; y<scr_h; y++,ty++) {
		for (tx = xo, x=0; x<scr_w; x++, tx++) {
			c = screen[k++];
			switch(c) {
				case 0:
					PUT_PIXEL(scr, x, y, MAKEPIX(' ', COLOR_GRAY));
					break; 
				case 1:
#ifdef ALT_DIRT
					hash = (abs(x+xo)^abs(y+yo))*2+x+y+xo+yo;
					PUT_PIXEL(scr, x, y, MAKEPIX(".,+:;`'+"[hash%7], PLASMA(tx, ty)<5?7:8));
#else
					
					PUT_PIXEL(scr, x, y, compute_plasma(tx, ty));
#endif
//					".,:;`'+"[hash%7], (((x+xo)*3^(y+yo))%3)?7:8));
					break;
				case 2: PUT_PIXEL(scr, x, y, MAKEPIX(
					".oOXOo."[(randtab[(ty*5001+tx)%4096]+(tick>>2))%7],BRIGHT(COLOR_GRAY)));
					break;
				case 3:	PUT_PIXEL(scr, x, y, MAKEPIX(
					"_.-~-~-."[(abs(xo+yo+x+y)+(tick>>1))%6], COLOR_BLUE));

					break;
				default: PUT_PIXEL(scr, x, y, MAKEPIX('?', 9));
			}
		}
	}
	//DEBUGDMP(screen, VIEWPNT_SIZE);
}

/* render all sprites */
void	render_sprites (struct viewpoint *vp)
{
	u16	i;
	/*
	if (vp->nobjs != 1) {
		printf("\nbagr\n");
		sleep(5);
	}
	*/
//	render_view(vp->view, vp->sx, vp->sy);
	for (i=0; i<vp->nobjs; i++) {
		u16 angle;
		u8 color;
		enum GFX gfxid;

		angle = vp->obj[i].angle;
		color = (angle>>6)&15;
		gfxid = (angle>>3)&7;
		angle &= 7;

		//DEBUG("angle %x color %x", vp->obj[i].angle, color);
		render_sprite(gfx_objects[gfxid], vp->obj[i].ox, vp->obj[i].oy, angle,
		              (gfxid == GFX_EXPL_ID || gfxid == GFX_BULLET_ID)?7:colors[color]);
	}
//	virtual_status(vp->en, vp->sh);
}


void	render_init(int w, int h)
{
	scr = malloc(w*h*sizeof(pixel));
	memset(scr, 0, w*h*sizeof(pixel));
	plasma_init();
}

void	screen_darken(int x, int y, int count, int color)
{
	int	i;
	for (i = 0; i < count; i++, x++) {
		pixel p = GET_PIXEL(scr, x, y);
		PUT_PIXEL(scr, x, y, MAKEPIX(PIXCHR(p), color));
	}
}

void	screen_write(int x, int y, int color, char *buf, int len)
{
	int	i;
	int	cur = color;
	for (i = 0; i < len; i++, x++) {
		unsigned char c = *buf++;
		if (c >= PRINTF_COLOR_ID) {
			if (c == PRINTF_COLOR_ID)
				cur = color;
			else
				cur = MAKEBG(COLBG(color))|(c-PRINTF_COLOR_ID);
			x--;
			continue;
		}
		if (c != ' ')
			PUT_PIXEL(scr, x, y, MAKEPIX(c, cur|MAKEBG(PIXBG(GET_PIXEL(scr, x, y)))));
	}
}

int	screen_printf(int x, int y, int color, char *fmt,...)
{
	va_list ap;
	static	char buf[1024];
	int	len;
	va_start(ap, fmt);
	screen_write(x, y, color, buf, len = vsprintf(buf, fmt, ap));
	va_end(ap);
	return len;
}


void	msg_add(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vsprintf(&msgs[msglast][0], fmt, ap);
	msgt[msglast] = tick;
	msglast++;
	if (msgshow < MSGS_MAX)
		msgshow++;
	msglast %= MSGS_MAX;
	va_end(ap);	
}

void	render_msgs()
{
	int	i;
	int	s;

	s = (MSGS_MAX + msglast - msgshow)%MSGS_MAX;
	for (i = 0; i < msgshow; i++,s++) {
		int c;

		s %= MSGS_MAX;
		if ((msgshow > MSGS_GOOD && tick - msgt[s] > FPS + FPS * MSGS_TIMEOUT / (msgshow - MSGS_GOOD + 1))
		    || tick - msgt[s] > FPS * MSGS_TIMEOUT) {
			i--; msgshow--;
			continue;
		}

		screen_darken(0, i, VIEWPNT_WIDTH, 8);
		c = COLOR_GREEN; /* green */
		if (tick-msgt[s] < 2)
			c = COLOR_DARKGRAY;
		if (tick-msgt[s] < 4)
			c = COLOR_LIGHTGREEN; /* lightgreen */
		screen_write(1, i, c, &msgs[s][0], strlen(&msgs[s][0]));
	}
}



void	editor_char(char ch)
{
	int len = strlen(talk_message);
	if (ch == '\r' || ch == '\n') {
		talk_mode = 0;
		return;
	}
	if (len > (VIEWPNT_WIDTH-10))
		return;
	if (ch == '\b' || ch == 127) {
		if (!len) {
			talk_mode = 0;
			return;
		}
		len--;
		talk_message[len] = 0;
		return;
	}
	talk_message[len++] = ch;
	talk_message[len] = 0;
		
}

void	render_editor()
{
	screen_darken(1, msgshow, VIEWPNT_WIDTH-4, COLOR_DARKGRAY);
	screen_printf(2, msgshow, COLOR_WHITE, "%c%s%c: %s%s", 
			 PRINTF_COLOR(plr_color(self_player->id)), self_player->nick, PRINTF_COLOR(0),
			talk_message, (tick&4)?"_":"");
}

void   render_pause()
{
	screen_darken((VIEWPNT_WIDTH-13)/2, (VIEWPNT_HEIGHT-1)/2-1, 13, COLOR_DARKGRAY);
	screen_darken((VIEWPNT_WIDTH-13)/2, (VIEWPNT_HEIGHT-1)/2,   13, COLOR_DARKGRAY);
	screen_darken((VIEWPNT_WIDTH-13)/2, (VIEWPNT_HEIGHT-1)/2+1, 13, COLOR_DARKGRAY);
	if ((tick * 2 / FPS) % 2)
		screen_printf((VIEWPNT_WIDTH-13)/2+1, (VIEWPNT_HEIGHT-1)/2, COLOR_LIGHTCYAN
				, "P A U S E D");
}


/* render everything */
void	render(struct viewpoint *vp)
{
	render_view(vp->view, vp->sx, vp->sy);
	render_sprites(vp);
	render_msgs();
	if (list_mode) {
		render_playerlist();
	} else {
		render_hud(vp);
	}
	if (talk_mode) render_editor();
	if (paused) render_pause();
//	render_ui();
	gio_scr_blit(scr);
}
