#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "common/buffer.h"
#include "common/debug.h"

struct buffer *buffer_create (int size, int for_reading)
{
	struct buffer *b;

	if (size == 0)
		return NULL;

	b = (struct buffer *) malloc( size + sizeof(struct buffer) - 1);
	b->p = b->n = 0;
	b->size = size;
	b->for_reading = for_reading;
	return b;
}

int	buffer_read (struct buffer *b, int fd)
{
	int result;

	if (!b->for_reading)
		return -1;

	result = read(fd, (char *)(b->p + b->data), b->size - b->p);
	if (result <= 0)
		return result;
	b->p += result;
	return result;
}

int	buffer_discard (struct buffer *b, int size)
{
	if (!b->for_reading)
		return -1;

	if (size > b->p)
		b->p = 0;
	else {
		b->p = b->p - size;
		memcpy(b->data, b->data + size, b->p);
	}

	return b->p;
}

/* tohleto by se dalo hooodne zoptimalizovat direct writem ale napadlo 
   me to az tedka a nechce se mi v tom zrovna hrabat, TODO! */
int	buffer_write (struct buffer *b, int fd, char *data, int len)
{
	if (b->for_reading)
		return -1;
	if ((len > b->size) || (b->p + len - b->n > b->size))
		return -2;
	if (b->p + len > b->size) {
		b->p -= b->n;
		memcpy(b->data, (char *)(b->data + b->n), b->p);
		b->n = 0;
	}
	memcpy((char *)(b->p + b->data), data, len);
	b->p += len;
	return buffer_write_flush(b, fd);
}

int 	buffer_write_flush (struct buffer *b, int fd)
{
	int written;
	if (b->for_reading)
		return -1;
	written = write(fd, (char *)(b->data + b->n), b->p - b->n);
	//DEBUG("WRITE f:%d l:%lu w:%d", fd, b->p - b->n, written);
	if (written == b->p - b->n) 
		b->p = b->n = 0;
	else if (written > 0)
		b->n += written;
	return (b->p - b->n);
}

void	buffer_destroy (struct buffer *b)
{
	if (b)
		free(b);
}

