#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "common/types.h"
#include "common/debug.h"


s16	bscanf(u8 *src, u16 len, u8 *fmt, ...)
{
	va_list ap;

	u8	*byte;
	u16	*word;
	u32	*dword;
	u16	lenreg=0, origlen=len;
	u16	*maxlen;

	va_start(ap, fmt);
	while (*fmt) {
		switch(*fmt++) {
			case 'b':
				if (len < 1)
					return -1;
				byte = va_arg(ap, u8 *);
				*byte = *src++;
				len--;
				continue;
			case 'w':
				if (len < 2)
					return -1;
				word = va_arg(ap, u16 *);
				*word = *src++;
				*word += (*src++)<<8;
				len -= 2;
				continue;
			case 'd':
				if (len < 4)
					return -1;
				dword = va_arg(ap, u32 *);
				*dword = *src++;
				*dword += (*src++)<<8;
				*dword += (*src++)<<16;
				*dword += (*src++)<<24;
				len -= 4;
				continue;
			case 'q':
				if (len < 2)
					return -1;

				maxlen = va_arg(ap, u16*);
				lenreg = *src++;
				lenreg += (*src++)<<8;
				len -= 2;
				if (lenreg >= *maxlen)
					return -2;
				*maxlen = lenreg;

				if (len < lenreg)
					return -1;
				byte = va_arg(ap, u8 *);
				while (lenreg--) {
					*byte++ = *src++;
					len--;
				}
				*byte = 0;
				continue;
		}
	}
	return (origlen - len);
}

u8	*bprintf(u8 *dest, u16 maxlen, u8 *fmt, ...)
{
	va_list ap;
	u16 	written=0;
	u8	byte, *string;
	u16	word;
	u32	dword;

	va_start(ap, fmt);
	while (*fmt) {
		switch(*fmt++) {
			case 'b':
				written++;
				if (written > maxlen)
					return NULL;
				byte = va_arg(ap, unsigned int);
				*dest++ = byte;
				continue;
			case 'w':
				written+=2;
				if (written > maxlen)
					return NULL;
				word = va_arg(ap, unsigned int);
				*dest++ = word&0xFF;
				*dest++ = (word>>8)&0xFF;
				continue;
			case 'd':
				written+=4;
				if (written > maxlen)
					return NULL;
				dword = va_arg(ap, u32);
				*dest++ = dword&0xFF;
				dword >>= 8;
				*dest++ = dword&0xFF;
				dword >>= 8;
				*dest++ = dword&0xFF;
				dword >>= 8;
				*dest++ = dword&0xFF;
				continue;
			case 's':
				string = va_arg(ap, u8*);
				word = strlen(string);
				goto string;
			case 'q':
				word = va_arg(ap, int);
				string = va_arg(ap, u8*);
string:
				if (2+word+written > maxlen)
					return NULL;
				*dest++ = word&0xFF;
				*dest++ = (word>>8)&0xFF;
				memcpy(dest, string, word);
				dest += word;
				continue;	
		}
	}
	va_end(ap);
	return dest;
}

#if 0

int main ()
{
	char buffer[1024];


}

#endif
