#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include "common/conf.h"
#include "common/types.h"

/* Everything is initially commented out in the stock .cfg files so that
 * when we change default value of something, everyone who didn't modify
 * that option gets the new default. */


static int
load_value(char type, char *rvalue, void **argp)
{
	if (type == 'i') { /* int array */
		**(int **) argp = atoi(rvalue);

	} else if (type == 'f') { /* float array */
		**(float **) argp = atof(rvalue);

	} else if (type == 's') { /* string array */
		if (*rvalue == '"') {
			char *e;

			++rvalue;
			e = strchr(rvalue, '"');
			if (e) *e = 0;
		}
		*(char **) argp = strdup(rvalue);

	} else {
		return -1; /* bad array type */
	}

	return 0;
}

int
vinit_cfg_vars(FILE * fp, char *names, char *types, va_list args)
{
	char ln[256];
	va_list arglist;

	while (fgets(ln, sizeof(ln), fp)) {
		char *rname, *rvalue, *p;
		char *namep, *typep;
		int i;

		rname = ln;
		/* Drop leading whitespace */
		while (isspace(rname[0]))
			rname++;
		/* Skip if blank line or comment */
		if (rname[0] == 0 || rname[0] == '#')
			continue;

		rvalue = strchr(rname, '=');
		/* Error if no = found */
		if (rvalue == NULL)
			return -1;

		p = rvalue;
		while (p > rname && isspace(p[-1]))
			p--;
		*p = 0;

		rvalue++;
		while (isspace(rvalue[0]))
			rvalue++;

		i = strlen(rvalue) - 1;
		if (i < 1)
			return -1;

		/* If '\n' is missing, input exceeded buffer; error return */
		if (rvalue[i] != '\n')
			return -1;
		rvalue[i] = 0;

		va_copy(arglist, args);

		/* init types and names and skip leading spaces */
		namep = names;
		typep = types;
		while (*namep == ' ')
			++namep;
		while (*typep == ' ')
			++typep;

		while (*typep) {
			void *argp = (void *) va_arg(arglist, void *);
			char name[40], type[2];

			i = strcspn(namep, " ");
			if (i > sizeof(name))
				return -1;
			memmove(name, namep, i);
			name[i] = 0;

			if (strcmp(name, rname) != 0) {
				namep += i;
				while (*namep == ' ')
					++namep;
				i = strcspn(typep, " ");
				typep += i;
				while (*typep == ' ')
					++typep;

				continue;
			}

			i = strcspn(typep, " ");
			if (i > sizeof(type))
				return -1;
			memmove(type, typep, i);

			if (type[0] == 'a') { /* array */
				struct dyn_array *tmp = argp;
				tmp->arr = realloc(tmp->arr, sizeof(tmp->arr) * (tmp->len + 1));

				if (load_value(type[1], rvalue, &tmp->arr[tmp->len]))
					return -1;

				tmp->len++;
			} else if (load_value(type[0], rvalue, &argp))
				return -1;

			break;
		}

		va_end(arglist);
	}

	return 0;
}

void
load_cfg_vars(char *fname, char *names, char *types, ...)
{
	/* Try ./fname, then confdir/fname and then try to load
	 * ~/.tunneler/fname */
	FILE *fp;
	char namebuf[256];
	va_list args;
	
	/* First load system-wide configuration */
	namebuf[0] = 0;
	strncat(namebuf, fname, 255);
	if (!(fp = fopen(namebuf, "r"))) {
		namebuf[0] = 0;
		strncat(namebuf, SYSCONFDIR, 255);
		strncat(namebuf, fname, 255);
		if (!(fp = fopen(namebuf, "r"))) {
			fprintf(stderr, "Cannot open %s neither %s - not fatal.",
				fname, namebuf);
		}
	}
	va_start(args, types);
	if (vinit_cfg_vars(fp, names, types, args)) {
		fprintf(stderr, "Error parsing %s - not fatal.\n", namebuf);
	}
	va_end(args);
	fclose(fp);
	
	/* Now try to load user's configuration. */
	namebuf[0] = 0;
	strncat(namebuf, getenv("HOME"), 255);
	strncat(namebuf, "/.tunneler/", 255);
	strncat(namebuf, fname, 255);
	if ((fp = fopen(namebuf, "r"))) {
		va_start(args, types);
		if (vinit_cfg_vars(fp, names, types, args)) {
			fprintf(stderr, "Error parsing %s - not fatal.",
				namebuf);
		}
		va_end(args);
		fclose(fp);
	}
}

