#include <error.h>
#include <errno.h>
#include <netdb.h>
#include <stdlib.h>
#include "common/conn.h"
#include "common/debug.h"

struct conn	*conn_create (struct sockaddr_in addr)
{
	struct conn *c;

	c = (struct conn *) malloc(sizeof(struct conn));
	if (!c) return NULL;
	c->sin = addr;
	c->fd = -1;
	c->in = c->out = NULL;
	c->state = CS_ERROR;
	return c;
}

void	conn_destroy (struct conn *c)
{
	if (!c) return;
	if (c->in != NULL) buffer_destroy(c->in);
	if (c->out != NULL) buffer_destroy(c->out);
	close(c->fd);
	free(c);
}

struct conn	*conn_connect (struct sockaddr_in addr)
{
	struct conn *c = conn_create(addr);

	c->fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (c->fd == -1) return c; 


	c->state = CS_CONNECTED;
	if (connect(c->fd, (struct sockaddr *)&(c->sin), sizeof(struct sockaddr_in)))
		c->state = CS_ERROR;

	SET_NONBLOCK(c->fd);
	return c;
}

struct conn	*conn_bind (struct sockaddr_in addr, int nlist)
{
	struct conn *c = conn_create(addr);
	int one=1;

	c->fd = socket(PF_INET, SOCK_STREAM, 0);
	if (c->fd == -1) return c; 

	setsockopt(c->fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
	if (bind(c->fd, (struct sockaddr *)&(c->sin), sizeof(struct sockaddr_in))) 
		return c; 
	if (listen(c->fd, nlist)) {
		close(c->fd);
		return c;
	}
	SET_NONBLOCK(c->fd);
	c->state = CS_LISTEN;
	return c;
}

struct conn	*conn_accept (struct conn *c)
{
	struct conn *nc;
	struct sockaddr_in addr;
	socklen_t len = sizeof(addr);
	int fd;

	if (!c || c->state != CS_LISTEN)
		return NULL;
	fd = accept(c->fd, (struct sockaddr *)&addr, &len);
	if (fd == -1)
		return NULL;

	nc = conn_create(addr);
	nc->fd = fd;
	nc->state = CS_CONNECTED;

	SET_NONBLOCK(nc->fd);
	return nc;
}


struct conn	*conn_bind_port (int port, int nlist)
{
	struct sockaddr_in addr;

	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(port);
	addr.sin_family = AF_INET;

	return conn_bind (addr, nlist);
}


in_addr_t
inet_resolve (char *str)
{
	in_addr_t s_addr = inet_addr(str);

	if (s_addr == INADDR_NONE) {
		struct hostent *hent = gethostbyname(str);

		if (hent) s_addr = ((struct in_addr *) hent->h_addr)->s_addr;
	}

	return s_addr;
}

struct conn	*conn_connect_resolv (char *str, int port)
{
	struct sockaddr_in addr;

	addr.sin_addr.s_addr = inet_resolve(str);
	addr.sin_port = htons(port);
	addr.sin_family = AF_INET;

	return conn_connect(addr);
}

int	conn_finalize_connect (struct conn *c)
{
	int val;
	socklen_t len = sizeof(val);

	getsockopt(c->fd, SOL_SOCKET, SO_ERROR, &val, &len);

	c->state = val ? CS_ERROR : CS_CONNECTED;
	return val;
}

void	conn_setbuff (struct conn *c, int inbuff, int outbuff)
{
	c->in = buffer_create(inbuff, 1);
	c->out = buffer_create(outbuff, 0);
}

