#include <assert.h>
#include <stdlib.h>
#include "common/debug.h"
#include "common/lists.h"
#include "common/object.h"

struct object objects;


static struct sprite *
coords_pick_sprite(enum GFX gfxid, int angle)
{
	assert(gfxid < GFX_MAX_ID);
	assert(angle < gfx_objects[gfxid]->n);
	return gfx_objects[gfxid]->directions[angle]->frames[0];
}

void	sprite_by_coords(int x, int y, int angle, enum GFX gfxid, struct spritepos *p)
{
	struct sprite *s;

	s = coords_pick_sprite(gfxid, angle);
	p->sx = x/XPOINTS_PERCHAR - s->center_x;
	p->sy = y/YPOINTS_PERCHAR - s->center_y;
	p->sprite = s;

	//DEBUG("x:%d y:%d angle:%d sx:%d sy:%d", x, y, angle, p->sx, p->sy);
}

void	get_vector(u16 angle, s16 length, struct object *obj)
{
	obj->vx = obj->vy = 0;

	if (angle % 2) {
		length = length*7/10;
		obj->vx = -(angle/4*2 - 1) * length;
		obj->vy = (((angle+2)%8)/4*2 - 1) * length;
	} else switch (angle) {
		case 0: obj->vy = -length; break;
		case 2: obj->vx =  length; break;
		case 4: obj->vy =  length; break;
		case 6: obj->vx = -length; break;
	}

	//DEBUG("angle:%d speed:%d ax:%d ay:%d", angle, add, *ax, *ay);
}

void	advance_object(struct object *obj)
{
	obj->x += obj->vx;
	obj->y += obj->vy;
}

void    sprite_border(struct sprite *s, u16 angle, int *sx, int *sy)
{
        *sx = -s->center_x;
        *sy = -s->center_y;

        switch (angle) {
		case 0: *sx += s->center_x;
			break;
		case 1: *sx += s->width-1;
			break;
		case 2: *sx += s->width-1;
			*sy += s->center_y;
			break;
		case 3: *sx += s->width-1;
			*sy += s->height-1;
			break;
		case 4: *sx += s->center_x;
			*sy += s->height-1;
			break;
		case 5: *sy += s->height-1;
			break;
		case 6: *sy += s->center_y;
			break;
		case 7: break;
        }
}


void	init_objects(void)
{
	init_list(objects);
}

void	add_object(struct object *obj)
{
	add_to_list_end(objects, obj);
}

void	del_object(struct object *obj)
{
	del_from_list(obj);
}
