#ifndef __INCLUDED_OBJECT_H
#define __INCLUDED_OBJECT_H

#include "common/types.h"
#include "common/sprite.h"
#include "common/lists.h"


#define XPOINTS_PERCHAR 8
#define YPOINTS_PERCHAR	16


/* "Hard" sprite information already mapped to the surface. */
struct spritepos {
	/* tohle vyplnuje convcoors */

	/* Sprite's upper left corner position on the surface
	 * (fine position / [XY]POINTS_PERCHAR). */
	s16 sx, sy;

	struct sprite *sprite;
};

struct object {
	LIST_HEAD(struct object);

	u16 x, y; /* virtual (fine, [pt]) position */
	s16 vx, vy; /* vectors */

	s8 color;
	int angle;
	enum GFX gfxid;

	struct spritepos spos;
};

extern struct object objects;


/* Fills spritepos according to given coordinates. */
void	sprite_by_coords(int x, int y, int angle, enum GFX gfxid, struct spritepos *p);
#define sprite_by_obj(obj_, spos_)     sprite_by_coords((obj_)->x, (obj_)->y, (obj_)->angle, (obj_)->gfxid, spos_)
#define obj_sync_sprite(obj_)          sprite_by_obj((obj_), &(obj_)->spos)
#define nxtsprite_by_obj(obj_, spos_)  sprite_by_coords((obj_)->x + (obj_)->vx, (obj_)->y + (obj_)->vy, (obj_)->angle, \
                                                        (obj_)->gfxid, spos_)
#define rotsprite_by_obj(obj_, spos_, angle_) \
                                       sprite_by_coords((obj_)->x, (obj_)->y, angle_, (obj_)->gfxid, spos_)

void	get_vector(u16 angle, s16 length, struct object *obj);
void	advance_object(struct object *obj); /* Moves it along the vector. */

void    sprite_border(struct sprite *s, u16 angle, int *sx, int *sy);

void	init_objects(void);
void	add_object(struct object *obj);
void	del_object(struct object *obj);

#endif
