#ifndef __INCLUDED_PACKET_H
#define __INCLUDED_PACKET_H

#include "common/types.h"

#define	PACKET_SIZE	2048

#define UNPACKW(s)	((s)[0]|((s)[1]<<8))
#define PACKW(s,n)	do { (s)[0] = (n)&0xff; (s)[1] = ((n)>>8)&0xff; } while(0);


/* TCP messages */

#define	CH_JOIN		0x01
#define CH_LEAVE	0x02
#define CH_SAY		0x03
#define CH_PAUSE	0x04
#define CH_TEAMUP	0x05

#define CH_MSGJOIN	0x81
#define CH_MSGLEAVE	0x82
#define	CH_MSGKILL	0x83
#define CH_MSGCHAT	0x84
#define CH_HELLO	0x85
#define	CH_MSGEDEATH	0x86
#define CH_MSGID	0x87
#define	CH_MSGGLAMOR	0x88
#define	CH_MSGTELEFRAG	0x89
#define	CH_MSGPAUSE	0x8A
#define	CH_MSGGSTYLE	0x8B
#define	CH_MSGWINNER	0x8C
#define	CH_MSGTEAMUP	0x8D
#define	CH_MSGTEAMINFO	0x8E
#define CH_MSGFLAGPICK	0x8F
#define CH_MSGFLAGDROP	0x90
#define CH_MSGFLAGRET	0x91
#define CH_MSGFLAGWIN	0x92


/* UDP messages */

#include "udp.h"

#define	HC_HELLO	0x01
#define HC_START	0x02
#define HC_UPDATE	0x04

#define HS_HELLO	0x81
#define HS_UPDATE	0x82
#define HS_END		0x83
#define HS_ERROR	0x84

struct udp_packet {
	struct uconn	from;
	u8	type;
	union {
		struct {
			u16 	curseq;
			u16	diffseq;
			u16 	size;
			u8 	*data;
		} su;
		struct {
			u16	ackseq;
			/* Bitmap:
			 * 0..2: direction
			 * 3..5: speed
			 * 7   : firing */
			u8	direction;
		} cu;
	};
};

int	udppck_send_hello (struct uconn *c, u16 id);
int	udppck_send_supdate (struct uconn *c, u16 curseq, u16 diffseq, u16 size, u8 *diff);
int	udppck_send_cupdate (struct uconn *c, u16 ack, u8 dir);
int	udppck_send_opcode (struct uconn *c, u8 opcode);
int	udppck_recv (struct uconn *server, struct udp_packet *pck);
u16	in_chksum (u16 *ipbuf, u32 iplen );

#endif
