#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include "common/types.h"
#include "common/sprite.h"
#include "common/debug.h"
#include "common/xmalloc.h"
#include "gameio/colors.h"

struct gfx_object *gfx_objects[GFX_MAX_ID];


void	dump_gfx(struct gfx_object *g);


#define SKIP_NEWLINES do { char c; while ((c = getc(fr)) == '\n') (*lineno)++; ungetc(c, fr); } while (0)

static struct sprite *
load_sprite(FILE *fr, unsigned char *fname, int *lineno)
{
	char	line[128], imglist[9] = "ta", *i, c;
	int	width = -1, height = -1, cx = -1, cy = -1, x, y;
	struct sprite 	*s;

	fscanf(fr, "%d %d %d %d", &width, &height, &cx, &cy);
	while ((c = getc(fr)) == ' ' || c == '\t');
	if (c && isalpha(c) && c != '\n') {
		i = imglist;
		do {
			*(i++) = c;
		} while (isalpha(c = getc(fr)) && (i - imglist) < 9);
		*i = '\0';
		ungetc(c, fr);
	}
	(*lineno)++;

	DEBUG("nacitam sprajt %d*%d center %d*%d, imidze %s", width, height, cx, cy, imglist);
	if (width < 0 || height < 0 || cx < 0 || cy < 0) {
		fprintf(stderr, "%s:%d: Parse error.\n", fname, (*lineno) - 1);
		exit(1);
	}

	s = (struct sprite *) ALLOC(sizeof(struct sprite)-1+(width*height)*sizeof(struct sprite_char));
	s->width = width;
	s->height = height;
	s->center_x = cx;
	s->center_y = cy;

	for (i = imglist; *i; i++) {
		SKIP_NEWLINES;
		for (y = 0; y < s->height && fgets(line, 127, fr); y++) {
			int l = strlen(line) - 1;

			(*lineno)++;
			//DEBUG("[%c]-> %s", *i, line);

			/* TODO: Optimizable. */
			for (x = 0; x < s->width; x++) {
				struct sprite_char *c = &s->sprite[width * y + x];

				switch (*i) {
				case 't':
					if (x >= l)
						c->text = ' ';
					else
						c->text = line[x];
					break;

				case 'a':
					if (x >= l)
						c->attr = SCA_TRANSPARENT;
					else
						switch (line[x]) {
						case '|': c->attr = SCA_NORMAL; break;
						case '@':
						case '#': c->attr = SCA_STANDOUT; break;
						case '.': c->attr = SCA_TRANSPARENT; break;
						case ':': c->attr = SCA_TRANSLUCENT; break;
						default:  c->attr = isupper(line[x]) ? SCA_BRIGHTCOLOR : SCA_DARKCOLOR;
							switch (tolower(line[x])) {
							case 'k': c->color = COLOR_BLACK; break;
							case 'r': c->color = COLOR_RED; break;
							case 'g': c->color = COLOR_GREEN; break;
							case 'y': c->color = COLOR_BROWN; break;
							case 'b': c->color = COLOR_BLUE; break;
							case 'm': c->color = COLOR_MAGENTA; break;
							case 'c': c->color = COLOR_CYAN; break;
							case 'h': c->color = COLOR_GRAY; break;

							default:
								fprintf(stderr, "%s:%d: Invalid attribute '%c'.\n",
								        fname, *lineno, line[x]);
								exit(1);
							}
						}
					break;

				default:
					fprintf(stderr, "%s:%d: Invalid image type '%c'.\n", fname, *lineno, *i);
					exit(1);
				}
			}
		}
	}

	return s;
}

static struct spriteset *
load_spritegroup(FILE *fr, unsigned char *fname, int *lineno)
{
	struct spriteset *sg;
	int 	n, i;

	fscanf(fr, "%d %*s", &n); (*lineno)++;
	SKIP_NEWLINES;

	sg = ALLOC(sizeof(struct spriteset) + (n - 1) * sizeof(struct sprite *));
	sg->n = n;
	for (i=0; i<n; i++) {
		sg->frames[i] = load_sprite(fr, fname, lineno);
	}
	return sg;
}

struct gfx_object	*load_gfx_object(char *name)
{
	FILE 	*fr;
	struct gfx_object 	*g;
	char 	filename[256];
	int 	n, i;
	int     lno = 1;
	int     *lineno = &lno;

	snprintf(filename, 255, "sprites/%s", name);
	fr = fopen(filename, "r");
	if (fr == NULL) {
		snprintf(filename, 255, "%s/sprites/%s", DATADIR, name);
		fr = fopen(filename, "r");
	}
	if (fr == NULL) {
		perror(filename);
		exit(EX_NOINPUT);
	}

	DEBUG("loading gfx obj named '%s'", name);
	fscanf(fr, "%d %*s", &n); (*lineno)++;
	SKIP_NEWLINES;

	g = ALLOC(sizeof(struct gfx_object) + (n - 1) * sizeof(struct spriteset *));
	g->n = n;
	for (i=0; i<n; i++) {
		g->directions[i] = load_spritegroup(fr, name, lineno);
	}
	fclose(fr);
	/* dump_gfx(g); */
	return g;
}

void    load_default_gfx(void)
{
	gfx_objects[GFX_TANK_ID]     = load_gfx_object("tank");
	gfx_objects[GFX_EXPL_ID]     = load_gfx_object("explosion");
	gfx_objects[GFX_FLAGTANK_ID] = load_gfx_object("flagtank");
	gfx_objects[GFX_BULLET_ID]   = load_gfx_object("bullet");
	gfx_objects[GFX_HOME_ID]     = load_gfx_object("home");
	gfx_objects[GFX_FORTRESS_ID] = load_gfx_object("fortress");
	gfx_objects[GFX_FLAG_ID]     = load_gfx_object("flag");
}



#if 0

void	dump_sprite(struct sprite *s)
{
	int i, j;
	for (j=0; j<s->height; j++) {
		for (i=0; i<s->width; i++) {
			struct sprite_char c = s->sprite[s->width*j+i];
			static const char cc[] = "krgybmch";
			printf("%c", c.text);
			if (c.attr == SCA_DARKCOLOR)
				printf("[%c]", cc[(int) c.color]);
			else if (c.attr == SCA_BRIGHTCOLOR)
				printf("[%c]", toupper(cc[(int) c.color]));
			else
				printf("[%d]", c.attr);
		}
		putchar('\n');
	}
}

void	dump_gfx(struct gfx_object *g)
{
	int i;
	DEBUG("dumping gfx obj with %d directions", g->n);
	for (i=0; i<g->n; i++) {
		int j;

		printf("-- direction %d\n", i);
		for (j=0; j<g->directions[i]->n; j++) {
			printf("-- frame %d\n", i);
			dump_sprite(g->directions[i]->frames[j]);
		}
	}
	printf("-- end of dump\n");
}

#endif
