#include "common/conn.h"
#include "common/udp.h"
#include "common/debug.h"
#include "common/xmalloc.h"

struct uconn	*udp_create (void)
{
	struct uconn *new;

	new = (struct uconn *) ALLOC(sizeof(struct uconn));
	memset(new, 0, sizeof(struct uconn));
	new->fd = -1;
	new->state = UCS_ERROR;
	return new;
}

void	udp_destroy(struct uconn *u)
{
	if (u->fd >= 0) close(u->fd);
	if (u) FREE(u);
}

struct uconn	*udp_bind(struct sockaddr_in saddr)
{
	struct uconn *new = udp_create();

	new->fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (new->fd == -1) return new;
	if (bind(new->fd, (struct sockaddr *) &saddr, sizeof(saddr)))
		return new;
	new->state = UCS_CLIENT;
	return new;
}


struct uconn	*udp_client(char *str, int port)
{
	struct uconn *n;
	struct sockaddr_in c_sin, s_sin;

	c_sin.sin_family = AF_INET;
	c_sin.sin_addr.s_addr = htonl(INADDR_ANY);
	c_sin.sin_port = htons(0);

	s_sin.sin_addr.s_addr = inet_resolve(str);
	s_sin.sin_port = htons(port);
	s_sin.sin_family = AF_INET;

	n = udp_bind(c_sin);

	if (n->state != UCS_ERROR)
		n->sin = s_sin;
	return n;
}

struct uconn	*udp_server(int port)
{
	struct uconn *n;
	struct sockaddr_in s_sin;

	s_sin.sin_addr.s_addr = INADDR_ANY;
	s_sin.sin_port = htons(port);
	s_sin.sin_family = AF_INET;
	n = udp_bind(s_sin);

	if (n->state != UCS_ERROR){
		n->sin = s_sin;
		n->state = UCS_SERVER;
	}
	return n;
}

void	udp_close (struct uconn *u)
{
	udp_destroy(u);
}

int	udp_recv (struct uconn *u, char *buf, int max, struct uconn *source)
{
	int len = sizeof(struct sockaddr_in), ret;
	ret = recvfrom(u->fd, buf, max, MSG_DONTWAIT, (struct sockaddr *)&(source->sin), &len);
	DEBUGDUMP("RECV", buf, ret);
	return ret;
}

int	udp_send (struct uconn *u, char *data, int len)
{
#ifdef UDP_RANDDROP
	if (random()%100 < UDP_RANDDROP) {
		DEBUG("DROPPING PACKET");
		return 1;
	} else {
		DEBUGDUMP("SEND", data,len);
		return sendto(u->fd, data, len, 0, (struct sockaddr *)&(u->sin), sizeof(struct sockaddr_in));
	}
#else
	DEBUGDUMP("SEND", data,len);
	return sendto(u->fd, data, len, 0, (struct sockaddr *)&(u->sin), sizeof(struct sockaddr_in));
#endif
}

struct uconn	*udp_dup (struct uconn *server, struct sockaddr_in sin)
{
	struct uconn *n = udp_create();
	n->sin = sin;
	n->fd = server->fd;
	n->state = UCS_CLIENT;
	return n;
}

int	udp_cmp(struct uconn *a, struct uconn *b)
{
	return (memcmp(&(a->sin), &(b->sin), sizeof(struct sockaddr ))==0);
}

int	udp_localport(struct uconn *a)
{
	struct sockaddr_in sin;
	int len = sizeof(sin);
	getsockname(a->fd, (struct sockaddr *)&sin, &len);
	return htons(sin.sin_port);
}
