#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include "common/debug.h"
#include "common/math.h"
#include "common/object.h"
#include "common/surface.h"
#include "common/viewpoint.h"

void	viewmove(int shiftx, int shifty, u8 *dst, u8 *src)
{
	int y;

	memset(dst, 1, VIEWPNT_SIZE);
	if (shifty > 0) {
		for (y = 0; y < (VIEWPNT_HEIGHT-shifty); y++) {
			if (shiftx > 0)
				memmove(shiftx + dst + (y+shifty) * VIEWPNT_WIDTH, 
				        src + y * VIEWPNT_WIDTH, VIEWPNT_WIDTH - shiftx);
			else
				memmove(dst + (y+shifty) * VIEWPNT_WIDTH,
					src - shiftx + y * VIEWPNT_WIDTH, VIEWPNT_WIDTH + shiftx);
		}
	} else {
		for (y = 0; y < (VIEWPNT_HEIGHT+shifty); y++) {
			if (shiftx > 0)
				memmove(shiftx + dst + y * VIEWPNT_WIDTH, 
				        src + (y-shifty) * VIEWPNT_WIDTH, VIEWPNT_WIDTH - shiftx);
			else
				memmove(dst + y * VIEWPNT_WIDTH,
					src - shiftx + (y-shifty) * VIEWPNT_WIDTH, VIEWPNT_WIDTH + shiftx);
		}
	}
}


/* fractal compression by sd@fucksheep.org */
#define MAXDEPTH 4096
int	diffstack[MAXDEPTH];
int	diffstack_ptr;
#define STACK_INIT() diffstack_ptr = MAXDEPTH;
#define PUSH(x) { assert(diffstack_ptr); diffstack[--diffstack_ptr] = x; }
#define POP() diffstack[diffstack_ptr++]
u32	*bitptr;	/* PNTer to bitmap */
int	bitsleft;	/* number of bits left */


#define INIT_BITS(ptr,w) \
{ \
	bitptr = (void *) ptr; \
	if (w) \
		*bitptr = 0; \
	bitsleft = 32; \
}
#define PUT_BITS(x,n) \
{ \
	if (bitsleft < n) { \
		*bitptr++ |= x & ((1<<bitsleft)-1); \
		*bitptr = 0; \
		*bitptr |= (x>>bitsleft) << (bitsleft = 32 - (n-bitsleft)); \
	} else { \
		*bitptr |= x << (bitsleft = bitsleft - n); \
	} \
}
#define GET_BITS(to, n) \
{ \
	if (bitsleft<n) { \
		to = (*bitptr & ((1<<bitsleft)-1)) | ((*(bitptr+1) >> (32 - (n-bitsleft))) << bitsleft); \
		bitptr++; \
		bitsleft = 32 - (n-bitsleft); \
	} else { \
		bitsleft -= n; \
		to = (*bitptr >> bitsleft) & ((1<<n)-1); \
	} \
}

#define PIXEL(buf,x,y) buf[x + y*VIEWPNT_WIDTH]

/* compare two regions, return 0 if same, 1 if not */
static	inline int cmp_region(u8 *a, u8 *b, int x, int y, int w, int h)
{
	int sy, ey;

	ey = y + h;
	for (sy = y; sy < ey; sy++)
		if (memcmp(&PIXEL(a, x, sy), &PIXEL(b, x, sy), w))
			return 1;
	return 0;
}

/*
 * 1|2
 *--|--
   3|4
 *
 */


/* fractal tree compression */
u8	*diff(u8 *old, u8 *new, int len, u8 *out, u8 *end)
{
	int x, y, w, h;
	/* stack emulation */
	STACK_INIT();
	/* bit output */
	INIT_BITS(out, 1);
	PUSH(0);
	PUSH(0); PUSH(0); PUSH(VIEWPNT_WIDTH); PUSH(VIEWPNT_HEIGHT);
	PUSH(1);

	while (1) {
		if (!POP())
			break;
		h = POP();
		w = POP();
		y = POP();
		x = POP();
		/* cannot be - ignore */
		if (!w || !h)
			continue;
		/* here we go :) */
		/* is the region same? */
		if (cmp_region(old, new, x, y, w, h)) {
//			printf("region %d %d, %d %d not same, old=%c new=%c\n", x, y, w, h, PIXEL(old, x, y), PIXEL(new, x, y));
			PUT_BITS(0, 1);
	
			/* did we reached the infinity? */
			if ((w == 1) && (h == 1)) {
				u32 pix = PIXEL(new, x, y);
				PUT_BITS(pix, 2);
				continue;
			}

			/* first quadrant */
			PUSH(x); PUSH(y); PUSH(w/2); PUSH(h/2); PUSH(1);
			/* second quadrant */
			PUSH(x + w/2); PUSH(y); PUSH(w-(w/2)); PUSH(h/2); PUSH(1);
			/* third quadrant */
			PUSH(x); PUSH(y + h/2); PUSH(w/2); PUSH(h-h/2); PUSH(1);
			/* fourth quadrant */
			PUSH(x + w/2); PUSH(y + h/2); PUSH(w-w/2); PUSH(h-h/2); PUSH(1);
			continue;
		}
		/* they're same, no data */
		PUT_BITS(1, 1);
//		printf("region %d %d, %d %d are same\n", x, y, w, h);
	}
	if (((u8 *) (bitptr + 1)) > end)
		return NULL;
	return (void *) (bitptr + 1);
}

u8	*patch(u8 *old, u8 *oend, u8 *patch, u8 *pend)
{
	int x, y, w, h;
	/* stack emulation */
	STACK_INIT();
	/* bit output */
	INIT_BITS(patch, 0);
	PUSH(0);
	PUSH(0); PUSH(0); PUSH(VIEWPNT_WIDTH); PUSH(VIEWPNT_HEIGHT);
	PUSH(1);
	
	while (1) {
		u32 bit;
		if (!POP())
			break;
		h = POP();
		w = POP();
		y = POP();
		x = POP();
		/* cannot be - ignore */
		if (!w || !h)
			continue;

		/* here we go :) */
		/* is the region same? */
		GET_BITS(bit, 1)
		if (!bit) {
//			printf("region %d %d, %d %d not same\n", x, y, w, h);
			/* did we reached the infinity? */
			if ((w == 1) && (h == 1)) {
				GET_BITS(bit, 2);
				PIXEL(old, x, y) = bit;
				continue;
			}

			/* first quadrant */
			PUSH(x); PUSH(y); PUSH(w/2); PUSH(h/2); PUSH(1);
			/* second quadrant */
			PUSH(x + w/2); PUSH(y); PUSH(w-(w/2)); PUSH(h/2); PUSH(1);
			/* third quadrant */
			PUSH(x); PUSH(y + h/2); PUSH(w/2); PUSH(h-h/2); PUSH(1);
			/* fourth quadrant */
			PUSH(x + w/2); PUSH(y + h/2); PUSH(w-w/2); PUSH(h-h/2); PUSH(1);
			continue;
		}
//		printf("region %d %d, %d %d are same\n", x, y, w, h);
	}
	return oend;
}



u8	*diff_view (struct viewpoint *old, struct viewpoint *new, u8 *output, u16 maxlen)
{
	struct viewpoint tmp;
	s32	xoff = new->sx - old->sx, yoff = new->sy - old->sy;
	u32 	i;
	u8 	*p;

	if (new->nobjs*4+3 > maxlen)
		return NULL;

	tmp = *old;
	DEBUG("xoffset: %ld yoffset: %ld nobjs: %d need: %d free: %d", xoff, yoff, new->nobjs, new->nobjs*3+3, maxlen);

	if (xoff >= VIEWPNT_WIDTH || yoff >= VIEWPNT_HEIGHT || xoff <= -VIEWPNT_WIDTH || yoff <= -VIEWPNT_HEIGHT) {
		xoff = 0x7f;
		yoff = 0x7f;
	} else {
		viewmove(-xoff, -yoff, tmp.view, old->view);
	}

	p = output;
	*p++ = xoff + 0x80;
	*p++ = yoff + 0x80;
	*p++ = new->sh;
	*p++ = new->en;
	*p++ = new->nobjs;
	for (i=0; i < new->nobjs; i++) {
		*p++ = new->obj[i].ox + 0x80;
		*p++ = new->obj[i].oy + 0x80;
		*p++ = new->obj[i].angle & 0xFF;
		*p++ = new->obj[i].angle >> 8;
	}

	//DEBUG("p:%x output:%x", p, output);
	p = diff(tmp.view, new->view, VIEWPNT_SIZE/4, p, p+maxlen-(3+2+4*new->nobjs));
	return p;

}

u16	patch_view (struct viewpoint *old, struct viewpoint *new, u8 *pstr, u16 len)
{
	u8	*p = pstr;
	s32 	xoff, yoff;
	u32	i;
	struct viewpoint *nv = new->next;
	u16 seq = new->seq;

	*new = *old;
	new->next = nv;
	new->seq = seq;
	if (len < 5)
		return -1;
	xoff = (*p++) - 0x80;
	yoff = (*p++) - 0x80;
	new->sh = *p++;
	new->en = *p++;

	if (xoff != 0x7f && yoff != 0x7f) {
 		new->sx = old->sx + xoff;
 		new->sy = old->sy + yoff;
	}
	new->nobjs = (*p++);
	if (len < 3+2+4*new->nobjs)
		return -1;
	if (new->nobjs > 16)
		return -1;
	for (i=0; i<new->nobjs; i++) {
		new->obj[i].ox = *p++ - 0x80;
		new->obj[i].oy = *p++ - 0x80;
		new->obj[i].angle = *p++;
		new->obj[i].angle |= *p++ << 8;
	}

	DEBUG("xoffset: %ld yoffset: %ld nobjs: %d need: %d free: %d", xoff, yoff, new->nobjs, new->nobjs*4+3, len);
	if (xoff != 0x7f && yoff != 0x7f) {
		viewmove(-xoff, -yoff, new->view, old->view);
	}

	if (patch(new->view, new->view+VIEWPNT_SIZE, p, pstr+len) == NULL) {
		DEBUG("OMG");
		return -1;
	}
	return 0;
}


void	calc_view(struct object *focusobj, struct viewpoint *v)
{
	s16 	i,j, sx, sy;
	u8	c;
	struct object *obj;

	memset(&(v->view), 0, VIEWPNT_SIZE);

	sx = focusobj->spos.sx-VIEWPNT_WIDTH/2;	
	sy = focusobj->spos.sy-VIEWPNT_HEIGHT/2;	
	for (j=0; j<VIEWPNT_HEIGHT; j++) {
		for (i=0; i<VIEWPNT_WIDTH; i++) {
			if (i+sx >= 0 && j+sy >= 0 && i+sx < SURFACE_WIDTH && j+sy < SURFACE_HEIGHT) {
				c = (surface[i+sx][j+sy]&(G_DIRT|G_STONE|G_HOME));
				v->view[i+j*VIEWPNT_WIDTH] = c==4?3:c;
			} else {
				v->view[i+j*VIEWPNT_WIDTH] = 0;
			}
		}
	}
	i = 0;

	foreach (obj, objects) {
		struct spritepos *s = &obj->spos;
		if (INTERSEC(sx, sy, VIEWPNT_WIDTH, VIEWPNT_HEIGHT,
		             s->sx, s->sy, s->sprite->width, s->sprite->height)) {
			v->obj[i].ox = s->sx - sx;
			v->obj[i].oy = s->sy - sy;
			v->obj[i].angle = ((obj->angle)&7) | (((obj->gfxid)&7)<<3) | (((obj->color)&15)<<6);
			//DEBUG("%x -> %x", obj->color, (obj->color&15)<<5);
			//DEBUG("x: %d y: %d state: 0x%x", v->obj[i].ox, v->obj[i].oy, v->obj[i].angle);
			if (++i >= 15)
				break;
		}
	}
	//DEBUG("nobjs: %d", i);
	v->nobjs = i;
	v->sx = sx;
	v->sy = sy;
}
