/*
 * tunneler game input (kbd) and output (scr) handling
 * these defs are common to all drivers
 * some of the rawkey stuff stolen from allegro game library -sd
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "gameio/gameio.h"

#if 0
/* common ascii mapping for rawkey */
static int common_ascii[RAWKEY_MAX] =
{
   0,   'a', 'b', 'c', 'd', 'e', 'f', 'g',
   'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
   'p', 'q', 'r', 's', 't', 'u', 'v', 'w',
   'x', 'y', 'z', '0', '1', '2', '3', '4',
   '5', '6', '7', '8', '9', '0', '1', '2',
   '3', '4', '5', '6', '7', '8', '9', 0,
   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   27,  0,   0,   0,   0,
   9,   0,   0,   13,  0,   0,   0,   0,
   0,   0,   0,   ' ', 0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   '/', '*',
   '-', '+', 0,  13,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   '=',
   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0
};
#endif


char *key_names[RAWKEY_MAX] =
{
   "(none)",     "A",          "B",          "C",
   "D",          "E",          "F",          "G",
   "H",          "I",          "J",          "K",
   "L",          "M",          "N",          "O",
   "P",          "Q",          "R",          "S",
   "T",          "U",          "V",          "W",
   "X",          "Y",          "Z",          "0",
   "1",          "2",          "3",          "4",
   "5",          "6",          "7",          "8",
   "9",          "PAD 0",      "PAD 1",      "PAD 2",
   "PAD 3",      "PAD 4",      "PAD 5",      "PAD 6",
   "PAD 7",      "PAD 8",      "PAD 9",      "F1",
   "F2",         "F3",         "F4",         "F5",
   "F6",         "F7",         "F8",         "F9",
   "F10",        "F11",        "F12",        "ESC",
   "~",          "-",          "=",          "BACKSPACE",
   "TAB",        "(",          ")",          "ENTER",
   ";",          "'",          "BACKSLASH",  "BACKSLASH2",
   ",",          ".",          "/",          " ",
   "INSERT",     "DEL",        "HOME",       "END",
   "PGUP",       "PGDN",       "LEFT",       "RIGHT",
   "UP",         "DOWN",       "PAD /",      "PAD *",
   "PAD -",      "PAD +",      "PAD DEL",    "PAD ENTER",
   "PRINT",      "PAUSE",      "KEY94",      "KEY95",
   "KEY96",      "KEY97",      "KEY98",      "KEY99",
   "KEY100",     "KEY101",     "KEY102",     "PAD =",
   "KEY104",     "KEY105",     "KEY106",     "KEY107",
   "KEY108",     "KEY109",     "KEY110",     "KEY111",
   "KEY112",     "KEY113",     "KEY114",     "LSHIFT",
   "RSHIFT",     "LCONTROL",   "RCONTROL",   "ALT",
   "ALTGR",      "LWIN",       "RWIN",       "MENU",
   "SCRLOCK",    "NUMLOCK",    "CAPSLOCK"
};

int	rawkeys[RAWKEY_MAX];
int	rawmods = 0;

char	*gio_error = NULL;
int	scr_w, scr_h;

static	int	autorepeat = 0;


int	gio_kbd_name2raw(char *name)
{
	int	i;
	for (i = 0; i < RAWKEY_MAX; i++) {
		if (!strcasecmp(key_names[i], name))
			return i;
	}
	return 0;
}

#define KBSIZE	32
static	int  keybuf[KBSIZE];
static	int  nkey;
static	int  raw_tick;

/* handle raw key press/release - this is called ONLY from internal keyboard driver
   (or if you just want to ungetc() on raw keyboard */
int	gio_handle_raw(int scancode, int ascii, int press)
{
	static	int  last_tick; /* when was the last key pressed */
	static	int  last_key; /* what key it was */

	/* someone seem to be typing on the keyboard */
	if (ascii > 0) do {
		if (last_key == ascii) {
			if (!press) {
				last_key = -1;
				break;
			}
			/* no repeatyet ?*/
			if (last_tick + autorepeat >= raw_tick)
				break;
		} else {
			if (!press)
				break;
		}
		/* setup tick till next repeat */
		last_tick = raw_tick;
		last_key = ascii;
		if (nkey < KBSIZE)
			keybuf[nkey++] = (scancode<<8) + ascii;
	} while (0);

	rawkeys[scancode] = press;
	return 0;
}

/*  get one key. tick is time counter for raw keyboard repeater delay */
int	gio_kbd_key(int tick)
{
	if (!kbd_is_raw)
		return kbd_driver->getkey();

	/* the driver have to call gio_handle_raw to add/remove keys */
	raw_tick = tick;
	kbd_driver->getkey();
	if (nkey)
		return keybuf[--nkey];
	return -1;
}

/*************************** PUBLIC STUFF *******************************/
/* The first driver in the arrays is the default one. */
struct	gi_scr_driver *scr_drivers[MAX_DRIVERS] = {
	&std_scr,
	&linux_scr,
	NULL
};

struct	gi_kbd_driver *kbd_drivers[MAX_DRIVERS] = {
	&std_kbd,
	&linux_kbd,
	NULL
};

struct	gi_scr_driver *scr_driver = NULL;
struct	gi_kbd_driver *kbd_driver = NULL;

int     gio_scr_find_driver(char *name)
{
	int	i;

	for (i = 0; scr_drivers[i]; i++) {
		scr_driver = scr_drivers[i];
		if (!strcasecmp(scr_driver->name, name)) {
			return i;
		}
	}

	return -1;
}

int     gio_kbd_find_driver(char *name)
{
	int	i;

	for (i = 0; kbd_drivers[i]; i++) {
		kbd_driver = kbd_drivers[i];
		if (!strcasecmp(kbd_driver->name, name)) {
			return i;
		}
	}

	return -1;
}

int	gio_scr_init(int id, int w, int h)
{
	scr_driver = scr_drivers[id];
	assert(scr_driver);
	if (scr_driver->init(w, h) < 0) {
		fprintf(stderr,"gio_scr_init(%d, %d, %d): %s\n",
			id, w, h, gio_error);
		return -1;
	}
	scr_w = w;
	scr_h = h;
	return 0;
}

int	gio_kbd_init(int id, int fd, int ar)
{
	kbd_driver = kbd_drivers[id];
	assert(kbd_driver);
	if (kbd_driver->init(fd) < 0) {
		fprintf(stderr, "gio_kbd_init(%d, %d): %s\n",
			id, fd, gio_error);
		return -1;
	}
	autorepeat = ar;
	return 0;
}

int	gio_scr_cleanup()
{
	return scr_driver->cleanup();
}

/* fun to not call this in raw keyboard mode */
int	gio_kbd_cleanup()
{
	return kbd_driver->cleanup();
}

/* blit pixel buffer on screen */
void	gio_scr_blit(pixel *buf)
{
	scr_driver->blit(buf);
}

