#ifndef __INCLUDED_GAMEIO_H
#define __INCLUDED_GAMEIO_H

/*
 * tunneler game input (kbd) and output (scr) handling
 * these defs are common to all drivers
 * some of the rawkey stuff stolen from allegro game library -sd
 */

#include "common/types.h"



/* Colors */

#include "gameio/colors.h"



/* Pixels */

typedef u16 pixel;

#define	MAKEPIX(chr,col)	(((col)<<8)|(chr))
#define PIXCOL(x)		((x)>>8)
#define PIXBG(x)		((x)>>12)
#define PIXFG(x)		(((x)>>8)&0xf)
#define	PIXCHR(x)		((x)&0xff)

#define PUT_PIXEL(buf, x, y, p) \
	buf[(y*scr_w)+(x)] = p;
#define GET_PIXEL(buf, x, y) \
	buf[(y*scr_w)+(x)]

#define GET_PIXEL_CHECK(buf, x, y) \
	(((y >= 0) && (x >= 0) && (x < scr_w) && (y < scr_h))?buf[y*scr_w+x]:0)
#define PUT_PIXEL_CHECK(buf, x, y, p) \
	if ((y >= 0) && (x >= 0) && (x < scr_w) && (y < scr_h)) \
		buf[y*scr_w+x] = p;



/* Keyboard */

/* raw keyboard mappings */
enum {
   RAWKB_SHIFT_FLAG    = 0x0001,
   RAWKB_CTRL_FLAG     = 0x0002,
   RAWKB_ALT_FLAG      = 0x0004,
   RAWKB_LWIN_FLAG     = 0x0008,
   RAWKB_RWIN_FLAG     = 0x0010,
   RAWKB_MENU_FLAG     = 0x0020,
   RAWKB_COMMAND_FLAG  = 0x0040,
   RAWKB_SCROLOCK_FLAG = 0x0100,
   RAWKB_NUMLOCK_FLAG  = 0x0200,
   RAWKB_CAPSLOCK_FLAG = 0x0400,
   RAWKB_INALTSEQ_FLAG = 0x0800,
};


enum {
   RAWKEY_A            = 1,
   RAWKEY_B            = 2,
   RAWKEY_C            = 3,
   RAWKEY_D            = 4,
   RAWKEY_E            = 5,
   RAWKEY_F            = 6,
   RAWKEY_G            = 7,
   RAWKEY_H            = 8,
   RAWKEY_I            = 9,
   RAWKEY_J            = 10,
   RAWKEY_K            = 11,
   RAWKEY_L            = 12,
   RAWKEY_M            = 13,
   RAWKEY_N            = 14,
   RAWKEY_O            = 15,
   RAWKEY_P            = 16,
   RAWKEY_Q            = 17,
   RAWKEY_R            = 18,
   RAWKEY_S            = 19,
   RAWKEY_T            = 20,
   RAWKEY_U            = 21,
   RAWKEY_V            = 22,
   RAWKEY_W            = 23,
   RAWKEY_X            = 24,
   RAWKEY_Y            = 25,
   RAWKEY_Z            = 26,
   RAWKEY_0            = 27,
   RAWKEY_1            = 28,
   RAWKEY_2            = 29,
   RAWKEY_3            = 30,
   RAWKEY_4            = 31,
   RAWKEY_5            = 32,
   RAWKEY_6            = 33,
   RAWKEY_7            = 34,
   RAWKEY_8            = 35,
   RAWKEY_9            = 36,
   RAWKEY_0_PAD        = 37,
   RAWKEY_1_PAD        = 38,
   RAWKEY_2_PAD        = 39,
   RAWKEY_3_PAD        = 40,
   RAWKEY_4_PAD        = 41,
   RAWKEY_5_PAD        = 42,
   RAWKEY_6_PAD        = 43,
   RAWKEY_7_PAD        = 44,
   RAWKEY_8_PAD        = 45,
   RAWKEY_9_PAD        = 46,
   RAWKEY_F1           = 47,
   RAWKEY_F2           = 48,
   RAWKEY_F3           = 49,
   RAWKEY_F4           = 50,
   RAWKEY_F5           = 51,
   RAWKEY_F6           = 52,
   RAWKEY_F7           = 53,
   RAWKEY_F8           = 54,
   RAWKEY_F9           = 55,
   RAWKEY_F10          = 56,
   RAWKEY_F11          = 57,
   RAWKEY_F12          = 58,
   RAWKEY_ESC          = 59,
   RAWKEY_TILDE        = 60,
   RAWKEY_MINUS        = 61,
   RAWKEY_EQUALS       = 62,
   RAWKEY_BACKSPACE    = 63,
   RAWKEY_TAB          = 64,
   RAWKEY_OPENBRACE    = 65,
   RAWKEY_CLOSEBRACE   = 66,
   RAWKEY_ENTER        = 67,
   RAWKEY_COLON        = 68,
   RAWKEY_QUOTE        = 69,
   RAWKEY_BACKSLASH    = 70,
   RAWKEY_BACKSLASH2   = 71,
   RAWKEY_COMMA        = 72,
   RAWKEY_STOP         = 73,
   RAWKEY_SLASH        = 74,
   RAWKEY_SPACE        = 75,
   RAWKEY_INSERT       = 76,
   RAWKEY_DEL          = 77,
   RAWKEY_HOME         = 78,
   RAWKEY_END          = 79,
   RAWKEY_PGUP         = 80,
   RAWKEY_PGDN         = 81,
   RAWKEY_LEFT         = 82,
   RAWKEY_RIGHT        = 83,
   RAWKEY_UP           = 84,
   RAWKEY_DOWN         = 85,
   RAWKEY_SLASH_PAD    = 86,
   RAWKEY_ASTERISK     = 87,
   RAWKEY_MINUS_PAD    = 88,
   RAWKEY_PLUS_PAD     = 89,
   RAWKEY_DEL_PAD      = 90,
   RAWKEY_ENTER_PAD    = 91,
   RAWKEY_PRTSCR       = 92,
   RAWKEY_PAUSE        = 93,
   RAWKEY_ABNT_C1      = 94,
   RAWKEY_YEN          = 95,
   RAWKEY_KANA         = 96,
   RAWKEY_CONVERT      = 97,
   RAWKEY_NOCONVERT    = 98,
   RAWKEY_AT           = 99,
   RAWKEY_CIRCUMFLEX   = 100,
   RAWKEY_COLON2       = 101,
   RAWKEY_KANJI        = 102,
   RAWKEY_EQUALS_PAD   = 103,  /* MacOS X */
   RAWKEY_BACKQUOTE    = 104,  /* MacOS X */
   RAWKEY_SEMICOLON    = 105,  /* MacOS X */
   RAWKEY_COMMAND      = 106,  /* MacOS X */
   RAWKEY_UNKNOWN1     = 107,
   RAWKEY_UNKNOWN2     = 108,
   RAWKEY_UNKNOWN3     = 109,
   RAWKEY_UNKNOWN4     = 110,
   RAWKEY_UNKNOWN5     = 111,
   RAWKEY_UNKNOWN6     = 112,
   RAWKEY_UNKNOWN7     = 113,
   RAWKEY_UNKNOWN8     = 114,

   RAWKEY_MODIFIERS    = 115,

   RAWKEY_LSHIFT       = 115,
   RAWKEY_RSHIFT       = 116,
   RAWKEY_LCONTROL     = 117,
   RAWKEY_RCONTROL     = 118,
   RAWKEY_ALT          = 119,
   RAWKEY_ALTGR        = 120,
   RAWKEY_LWIN         = 121,
   RAWKEY_RWIN         = 122,
   RAWKEY_MENU         = 123,
   RAWKEY_SCRLOCK      = 124,
   RAWKEY_NUMLOCK      = 125,
   RAWKEY_CAPSLOCK     = 126,

   RAWKEY_MAX          = 127
};

/* pressed-keys map */
extern	int	rawkeys[RAWKEY_MAX];
extern	int	rawmods;

#define kbd_is_raw kbd_driver->raw



/* GIO itself */

#define GIO_ERROR(x)		{ gio_error = x; return -1; }

/* bone for screen driver */
struct	gi_scr_driver {
	char	*name;
	char	*desc;
	int	(*init)(int w, int h);
	int	(*cleanup)();
	void	(*blit)(pixel *buffer);
};

/* bone for keyboard driver */
struct	gi_kbd_driver {
	char	*name;
	char	*desc;
	int	raw;		/* is the driver raw ? (can have more keypresses at a time) */
	int	(*init)(int fd);
	int	(*cleanup)();
	int	(*getkey)();
};

#define MAX_DRIVERS 16

struct	gi_scr_driver linux_scr;
struct	gi_scr_driver std_scr;
struct	gi_kbd_driver std_kbd;
struct	gi_kbd_driver linux_kbd;
struct	gi_scr_driver *scr_drivers[MAX_DRIVERS];
struct	gi_kbd_driver *kbd_drivers[MAX_DRIVERS];
extern	char	*gio_error;
extern	int	scr_w, scr_h;

extern	int	gio_handle_raw(int scancode, int ascii, int press);

extern  int     gio_scr_find_driver(char *name);
extern  int     gio_kbd_find_driver(char *name);

extern	int	gio_scr_init(int id, int w, int h);
extern	int	gio_scr_cleanup();
extern	void	gio_scr_blit(pixel *buf);

extern	int	gio_kbd_init(int id, int fd, int ar);
extern	int	gio_kbd_key(int tick);
extern	int	gio_kbd_cleanup();
extern	int	gio_kbd_name2raw(char *name);

extern	struct	gi_scr_driver *scr_driver;
extern	struct	gi_kbd_driver *kbd_driver;

#endif
