/*
 * this driver uses linux /dev/vcsa* direct textmode access, fastest
 * possible option though you need to tweak perms (and sit behind physical terminal as well)
 * -sd
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <string.h>
#include "gameio/gameio.h"

static	int	vcsa_fd, /*scr_w, scr_h,*/ vcs_w, vcs_h;
static	pixel	*dbuf;

static	void	set_cursor(int mode)
{
	if (mode) 
		printf ("\033[?25h");
	 else
		printf ("\033[?25l");
	fflush (stdout);
}
/********************************** PUBLIC ******************************/
static	int	init(int w, int h)
{
	struct	stat st;
	int	major, minor;
	char	name[64];
	struct { unsigned char lines, cols, x, y; } scrn = { 0, 0, 0, 0 };

	if (fstat(2, &st) < 0)
		GIO_ERROR("cannot examine your console");
	
	major = st.st_rdev >> 8;
	minor = st.st_rdev & 0xff;
	
	if ((major != 4) || (minor >= 64))
		GIO_ERROR("this does not look like a linux virtual terminal");
	sprintf(name, "/dev/vcsa%d", minor);
	vcsa_fd = open(name, O_RDWR);
	if (vcsa_fd < 0)
		GIO_ERROR("cannot open vcsa; check perms");
	read(vcsa_fd, &scrn, 4);

	if ((scrn.cols < w) || (scrn.lines < h)) {
		close(vcsa_fd);
		GIO_ERROR("sorry, but you terminal is smaller than size requested");
	}

	scr_w = w;
	scr_h = h;
	vcs_w = scrn.cols;
	vcs_h = scrn.lines;
	dbuf = malloc(vcs_w * vcs_h * sizeof(pixel));
	memset(dbuf, 0, vcs_w * vcs_h * sizeof(pixel));
	printf("\e7");
	set_cursor(0);

	/* clear the screen */
	lseek(vcsa_fd, 4, SEEK_SET);
	memset(dbuf, 0, vcs_w * vcs_h * sizeof(pixel));
	write(vcsa_fd, dbuf, vcs_w * vcs_h * sizeof(pixel));

	return 0;
}

static	int	cleanup()
{
	free(dbuf);
	close(vcsa_fd);
	printf("\e8");
	set_cursor(1);
	fflush(stdout);
	return 0;
}

static	void	blit(pixel *buf)
{
	int	i;
	lseek(vcsa_fd, 4, SEEK_SET);
	/* we can blit directly */
	if (scr_w == vcs_w) {
		write(vcsa_fd, buf, scr_w * scr_h * sizeof(pixel));
	} else {
		/* we need to convert it to blit buffer */
		for (i = 0; i < scr_h; i++)
			memcpy(dbuf + i*vcs_w, buf + i*scr_w, scr_w*sizeof(pixel));
		write(vcsa_fd, dbuf, vcs_w * scr_h * sizeof(pixel));
	}
}

/* driver descriptor */
struct	gi_scr_driver linux_scr = {
	.name	=	"linux",
	.desc	=	"Linux /dev/vcsa* device (need perms!)",
	.init	=	init,
	.cleanup=	cleanup,
	.blit	=	blit
};

