/*
 * quite basic ansi output handler, this is last-resort option, because
 * anything else seem to perform better -sd
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "gameio/gameio.h"

#define CLRSCR	"\e[2J"
#define CLREOL	"\e[1K"
#define GOTOXY  "\e[%d;%df"
#define SETFGCOLOR "\e[3%dm"
#define SETBGCOLOR "\e[4%dm"
#define	SETBOLD "\e[1m"
#define SETNOBOLD "\e[0m"
#define TERM_STORE "\e7"
#define TERM_RESTORE "\e8"
#define HIDE_CURSOR "\e[?25l"
#define SHOW_CURSOR "\e[?25h"
#define	MAXOUTBUF	16384

/* ibm to ansi translator */
#define IBM2ANSI(c) (ibm2ansi[c&7]|(c&8))
static	int ibm2ansi[] = {0, 4, 2, 6, 1, 5, 3, 7};

static	pixel	*dbuf;
//static	int	scr_w, scr_h;
static	int	fgcolor = 7;
static	int	bgcolor = 0;
static	int	cfgcolor = 7;
static	int	cbgcolor = 0;
static	int	have_bold = 0;
static	int	curx, cury;
static	char	outbuf[MAXOUTBUF];
static	int	outbuflen = 0;

static	inline void	v_flush()
{
	write(1, outbuf, outbuflen);
	outbuflen = 0;
}

static	inline	void	v_write(char *buf, int len)
{
	if (len + outbuflen > (MAXOUTBUF-1))
		v_flush();
	memcpy(outbuf + outbuflen, buf, len);
	outbuflen += len;
}


static	inline void	v_goto(int x, int y)
{
	static	char	buf[512];

	v_write(buf, sprintf(buf, GOTOXY, y+1, x+1));
}

static inline void	v_bold(int b)
{
	if (b) {
		if (!have_bold) {
			v_write(SETBOLD, strlen(SETBOLD));
			have_bold = 1;
			cfgcolor = cbgcolor = -1;
		}
	} else {
		if (have_bold) {
			v_write(SETNOBOLD, strlen(SETNOBOLD));
			have_bold = 0;
			cfgcolor = cbgcolor = -1;
		}
	}
}

static inline void	v_sync_color()
{
	int	t;
	static	char	buf[128];
	static	char	buf2[128];

	buf[0] = buf2[0] = 0;

	v_bold(fgcolor & 8);
	t = fgcolor & 7;

	if (t != cfgcolor) {
		sprintf(buf, SETFGCOLOR, t);
		cfgcolor = t;
	}
	if (bgcolor != cbgcolor) {
		sprintf(buf2, SETBGCOLOR, bgcolor);
		cbgcolor = bgcolor;
	}
	v_write(buf, strlen(buf));
	v_write(buf2, strlen(buf2));
}

/********************************** PUBLIC ******************************/
int	init(int w, int h)
{

	curx = cury = 0;

	v_write(TERM_STORE, strlen(TERM_STORE));
	v_write(CLRSCR, strlen(CLRSCR));
	v_write(HIDE_CURSOR, strlen(HIDE_CURSOR));
	v_flush();

	dbuf = malloc(w*h*sizeof(pixel));
	scr_w = w;
	scr_h = h;
	return 0;
}

int	cleanup()
{
	free(dbuf);
	curx = cury = 0;
	v_write(TERM_RESTORE, strlen(TERM_RESTORE));
	v_write(SHOW_CURSOR, strlen(SHOW_CURSOR));
	v_flush();
	return 0;
}


void	blit(pixel *buf)
{
	int	x, y, pos;

	v_goto(0, 0);
	for (y = 0; y < scr_h; y++) {
		int width = (y == (scr_h-1))?scr_w-1:scr_w;
		x = 0;
retry:
		while ((x < width) && (buf[y*scr_w+x] == dbuf[y*scr_w+x]))
			x++;
		if (x >= width)
			continue;
		v_goto(x, y);
		while ((x < width) && (buf[y*scr_w+x] != dbuf[y*scr_w+x])) {
			pos = y*scr_w+x;
			fgcolor = IBM2ANSI(PIXFG(buf[pos]));
			bgcolor = IBM2ANSI(PIXBG(buf[pos]));
			v_sync_color();
			if (outbuflen > MAXOUTBUF)
				v_flush();
			outbuf[outbuflen++] = PIXCHR(buf[pos]);
			dbuf[pos] = buf[pos];
			x++;
		}
		goto retry;
	}
	v_flush();
}


/* driver descriptor */
struct	gi_scr_driver std_scr = {
	.name	=	"std",
	.desc	=	"standard ANSI terminal output",
	.init	=	init,
	.cleanup=	cleanup,
	.blit	=	blit
};
