#ifndef __INCLUDED_AI_H
#define __INCLUDED_AI_H

#include "server/game.h"

struct player;

/* struct ai life:
 * AI_IDLE:    The bot has nothing to do, staying on a single spot.
 *             It does so until idle_count is zero (it relaxes a little
 *             after it kills an enemy, more realistic) and then
 *             until it has full energy if it finds inself in a home
 *             and until it has full shields if it finds itself in
 *             own home. Then it switches to the AI_EXPLORE state.
 * AI_EXPLORE: The bot plays an explorer. It follows either randomly
 *             laid route or already digged paths, until it sees
 *             any enemy (->AI_HUNT).
 * AI_HUNT:    The bot pursues a chosen enemy and fires at will,
 *             until it kills the victim (->AI_IDLE for a short period).
 * AI_FLEE:    The bot got to low energy and/or shields.
 *             It heads to the nearest home (if energy) or to its own
 *             home (if shields). Then it goes to AI_IDLE.
 */

struct ai {
	enum ai_state {
		AI_IDLE,
		AI_EXPLORE,
		AI_HUNT,
		AI_FLEE,
		AI_MAX
	} state;
	union {
		struct {
			int tick_count;
		} idle;
		struct {
			/* How long to hold direction. */
			int tick_count;
		} explore;
		struct {
			struct player *pursuing;
		} hunt;
		struct {
			int home_x, home_y;
		} flee;
	};

	/* Backing off is a temporary state when the bot experienced trouble
	 * during turning around - this state kicks in for a few frames and
	 * tries to head in a random opposite direction. Then the normal
	 * state handler takes over again. */
	/* We try to get to a certain distance from those coordinates. */
	/* TODO: This is enough for simple stuff, but not if we got entangled
	 * in a more complicated structure we need to find a way out from.
	 * That would probably need a completely different algorithm with
	 * own state etc. (Follow the stone until there is a hole in it.)
	 * --pasky */
	int backing_off; /* How far away to back off (in special units);
	                  * 0 means no backing off. */
	int backoff_x, backoff_y;

	/* How many ticks ago did we get shot the last time? */
	long last_shot;
	/* How many ticks ago did we change angle the last time? */
	long last_turnaround;
};

/* First goes always the hooked struct ai. */
void    hit_ai_hook (struct player *victim, struct player *trespasser);
void    stuck_ai_hook (struct player *victim);
void    kill_ai_hook (struct player *killer, struct player *victim);
void    death_ai_hook (struct player *victim, struct player *killer);

void    reset_ai (struct ai *ai);
void    stop_ai_pursue (struct player *p);

void    update_bot (struct tank *tank);

#endif
