#ifndef __INCLUDED_GAME_H
#define __INCLUDED_GAME_H

#include "common/globgame.h"

#include "common/types.h"
#include "common/object.h"
#include "common/sprite.h"
#include "common/viewpoint.h"
#include "common/lists.h"

struct home;
struct player;


struct tank {
	LIST_HEAD(struct tank);

	enum tank_status {
		TANK_INIT = 0, /* !player->initialized; calloc() sets us this */
		TANK_READY, /* ready and well */
		TANK_FLAGGED, /* ready and carrying a flag */
		TANK_EXPLODING, /* a huge fireball envelopes... */
		TANK_PAUSED, /* the player has paused the game */
		TANK_OBSERVER, /* an immaterial observing tank; for last man standing */
	}       status;

	struct object obj;

	s16 	energy, shields;

	u8	color;
	s16	next_angle;
	u16	speed, dirt_points, dirt_penalty;

	u16	fire_burst;
	u8	firing, fire;
	int	bullets_in_flight, bullet_penalty;
	int	flag_team; /* team_id of the flag if carrying it */
	int	loser; /* tank lose the game; after exploding is over, switch to observer */

	/* TODO: This should be probably tracked in struct player instead. --pasky */
	int	score, kills_in_line;
	int	deaths;

	struct player *player;
	struct home *home;
};

#define tank_ready(tank_) ((tank_)->status == TANK_READY \
                        || (tank_)->status == TANK_FLAGGED)
#define tank_solid(tank_) ((tank_)->status == TANK_READY \
			|| (tank_)->status == TANK_FLAGGED \
			|| (tank_)->status == TANK_EXPLODING \
			|| (tank_)->status == TANK_PAUSED)

struct bullet {
	LIST_HEAD(struct bullet);

	struct object obj;
	s16	power;
	struct tank *owner;

	int refcount;
};

struct home {
	LIST_HEAD(struct home);
	struct object obj;
	struct player *player;
};

struct fortress {
	struct object obj;
	int team_id;
};

struct flag {
	struct object obj;
	int team_id;
};

struct npc {
	struct object obj;
	u16 speed, dir;
};


void    pause_tank (struct tank *t);
void    unpause_tank (struct tank *t);
void	explode_tank (struct tank *t);
int     tank_in_region(struct tank *tank, int x, int y, int wx, int wy);
struct player *find_player_by_xy (u16 x, u16 y, u16 wx, u16 wy);

struct flag *synth_flag (int x, int y, int team_id);

void	init_player (struct player *p, u32 id);
void	finit_player (struct player *p);
void	player_move (struct player *p, u8 direction, u8 speed, u8 firing);

void	init_game (void);
void	init_ctf_team (int team_id);
void    move_game (void);

void	tank_view (struct tank *t, struct viewpoint *v);

#endif
