#ifndef __INCLUDED_PLAYERS_H
#define __INCLUDED_PLAYERS_H

#include <stdlib.h>

#include "common/conn.h"
#include "common/lists.h"
#include "common/state.h"
#include "common/types.h"
#include "common/udp.h"
#include "server/ai.h"
#include "server/game.h"

/* Players */

struct player {
	LIST_HEAD(struct player);

	u16	id;
	u16	team;
	u16	initialized;

	u8	*name;
	u16	fps;

	struct tank	*tank;
	struct home	*home;
	struct ai	ai; /* TODO: union with some other struct containing udp,tcp,seqs,... */

	struct conn 	*tcp;
	struct uconn	*udp;

	u16	cur_seq, last_seq;
	struct stq	state;

	u8	current_direction;
};
#define player_is_ai(player) (player->udp == NULL && player->tcp == NULL)

extern struct dyn_array bot_names;
extern struct player players;
extern int num_players;

void    kill_tank(struct tank *t, struct tank *by);

void    init_players(void);

struct player	*alloc_client (struct conn *tcp);
struct player   *alloc_bot (void);
void	join (struct player *p);
void    kick(struct player *plr);
void    spawn_bot (void);
void    check_winner(void);



/* Teams */

struct team {
	int active;
	char *name;
	int players;

	/* Deathmatch stuff */
	int score, deaths;

	/* CtF stuff */
	int flost, fwon; /* flags lost, flags won */
	unsigned int fstolen:1;
};
/* Team 0 is special fallback team for players who didn't join a specific
 * team. Statistics aren't kept for it etc. */
extern struct team teams[MAX_TEAMS];

void    init_teams(void);
void    join_team(struct player *p, int teamid);

#endif
