#define HAVE_GETOPT_LONG /* XXX: Autodetect. Autoconf. F33r. --pasky */

#include <assert.h>
#include <getopt.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sysexits.h>
#include <time.h>
#include "common/bytef.h"
#include "common/conf.h"
#include "common/debug.h"
#include "common/packet.h"
#include "common/surface.h"
#include "common/xmalloc.h"
#include "server/console.h"
#include "server/players.h"
#include "server/server.h"
#include "version.h"

/**********/

struct uconn	*server;
struct conn	*chatserver;
struct	timeval tv;

u16 	objcounter=0;
u32	tick=0;

/* Options */
int     do_ferda = 0;
int     bots_to_spawn = 0;
#define DEFAULT_LMS_LIVES 3
int     lms_lives = DEFAULT_LMS_LIVES;
#define DEFAULT_FORCED_TEAMS 2
int     teams_enforce = 0;
int     teams_ai = 0;
enum game_style gstyle = GSTYLE_DM;

struct gfx_object	*gfx_tank, *gfx_bullet, *gfx_home, *gfx_explosion;


/**********/

void	unicast(struct player *p, u8 *msg, u16 len)
{
	write(p->tcp->fd, msg, len);
}
void	broadcast(u8 *msg, u16 len)
{
	struct player *p;
	foreach(p, players)
		if (p->initialized && !player_is_ai(p)) 
			unicast(p, msg, len);
}

void	bcast_kill(struct player *recv, struct player *send)
{
	u8	buffer[6];

	CONSOLE("Recorded frag of %d::%s by %d::%s", recv->id, recv->name, send->id, send->name);

	DEBUG("::: BCAST_KILL %d %d", recv->id, send->id);
	bprintf(buffer, 6, "www", CH_MSGKILL, recv->id, send->id);
	broadcast(buffer, 6);
}

void	bcast_telefrag(struct player *recv, struct player *send)
{
	u8	buffer[6];

	assert(recv != send);

	CONSOLE("Recorded telefrag of %d::%s by %d::%s", recv->id, recv->name, send->id, send->name);

	DEBUG("::: BCAST_TELEFRAG %d %d", recv->id, send->id);
	bprintf(buffer, 6, "www", CH_MSGTELEFRAG, recv->id, send->id);
	broadcast(buffer, 6);
}

void	bcast_energydeath(struct player *plr)
{
	u8	buffer[6];

	CONSOLE("Recorded energetic death of %d::%s", plr->id, plr->name);

	DEBUG("::: BCAST_EDEATH %d", plr->id);
	bprintf(buffer, 4, "ww", CH_MSGEDEATH, plr->id);
	broadcast(buffer, 4);
}

void	bcast_glamorize(struct player *plr, int level)
{
	u8	buffer[6];

	CONSOLE("%d::%s's legend reached level %d!", plr->id, plr->name, level);

	DEBUG("::: BCAST_GLAMOR %d %d", plr->id, level);
	bprintf(buffer, 6, "www", CH_MSGGLAMOR, plr->id, level);
	broadcast(buffer, 6);
}

void	bcast_pause(struct player *plr, int pause)
{
	u8	buffer[6];

	CONSOLE("%d::%s's pause state change to %d.", plr->id, plr->name, pause);

	DEBUG("::: BCAST_PAUSE %d %d", plr->id, pause);
	bprintf(buffer, 6, "www", CH_MSGPAUSE, plr->id, pause);
	broadcast(buffer, 6);
}

void	bcast_winner(struct player *plr)
{
	u8	buffer[6];

	CONSOLE("THE WINNER OF THE GAME IS %d::%s !", plr->id, plr->name);

	DEBUG("::: BCAST_WINNER %d", plr->id);
	bprintf(buffer, 4, "ww", CH_MSGWINNER, plr->id);
	broadcast(buffer, 4);
}

void	bcast_join(struct player *p)
{
	struct player	*q;
	u8	msg1[MAX_NICKLEN+10], msg2[MAX_NICKLEN+10];
	u16	len1=strlen(p->name), len2;
	int i;

	CONSOLE("Connecting player %d::%s on %d FPS (team %d::%s)",
	        p->id, p->name, p->fps, p->team, teams[p->team].name);

	if (!player_is_ai(p)) {
		/* First send info about teams. */
		for (i = 1; i < MAX_TEAMS; i++) {
			char buf[MAX_NICKLEN + 16];
			int namelen = strlen(teams[i].name);

			if (!teams[i].active)
				continue;
			bprintf(buf, MAX_NICKLEN + 16, "wwwwqwww",
			      CH_MSGTEAMINFO, i,
			      teams[i].score, teams[i].deaths,
			      namelen, teams[i].name,
			      teams[i].fwon, teams[i].flost, teams[i].fstolen);
			unicast(p, buf, namelen + 16);
		}
	}

	DEBUG("::: BCAST_JOIN @%d as '%s'", p->id, p->name);
	bprintf(msg1, 42, "wwwwq", CH_MSGJOIN, p->id, p->tank->score, p->tank->deaths, len1, p->name);
	foreachback (q, players) { /* ...in the order of appearance */
		if (q->initialized) {
			if (p != q && !player_is_ai(p)) {
				len2 = strlen(q->name);
				bprintf(msg2, 42, "wwwwq",
				      CH_MSGJOIN, q->id, q->tank->score, q->tank->deaths, len2, q->name);
				unicast(p, msg2, len2+10);
				if (q->team) {
					u8 buffer[6];

					bprintf(buffer, 6, "www", CH_MSGTEAMUP, q->id, q->team);
					unicast(p, buffer, 6);
				}
				if (q->tank->status == TANK_PAUSED) {
					u8 buffer[6];

					bprintf(buffer, 6, "www", CH_MSGPAUSE, q->id, 1);
					unicast(p, buffer, 6);
				}
			}
			if (!player_is_ai(q)) {
				unicast(q, msg1, len1+10);
			}
		}
	}
}

void	bcast_leave(struct player *p)
{
	char	buffer[8];

	CONSOLE("Disconnecting player %d::%s", p->id, p->name);

	DEBUG("::: BCAST_LEAVE %d", p->id);
	bprintf(buffer, 4, "ww", CH_MSGLEAVE, p->id);
	broadcast(buffer, 4);
}

void	bcast_say(struct player *p, u8 *buf, u16 len)
{
	char	buffer[MAX_MSGLEN+6];

	CONSOLE("<%d::%s> %.*s", p->id, p->name, len, buf);

	bprintf(buffer, MAX_MSGLEN+6, "wwq", CH_MSGCHAT, p->id, len, buf);
	DEBUG("::: BCAST_SAY '%d'", p->id);
	broadcast(buffer, len+6);
}

void	bcast_teamup(struct player *p, int teamid)
{
	char	buffer[6];

	CONSOLE("%d::%s joins team %d::%s", p->id, p->name, teamid, teams[teamid].name);

	bprintf(buffer, 6, "www", CH_MSGTEAMUP, p->id, teamid);
	DEBUG("::: BCAST_TEAMID %d %d", p->id, teamid);
	broadcast(buffer, 6);
}

void	bcast_flagpick(struct player *p, int teamid)
{
	char	buffer[6];

	CONSOLE("%d::%s picks a flag of team %d::%s", p->id, p->name, teamid, teams[teamid].name);

	bprintf(buffer, 6, "www", CH_MSGFLAGPICK, p->id, teamid);
	DEBUG("::: BCAST_FLAGPICK %d %d", p->id, teamid);
	broadcast(buffer, 6);
}

void	bcast_flagdrop(struct player *p, int teamid)
{
	char	buffer[6];

	CONSOLE("%d::%s drops a flag of team %d::%s", p->id, p->name, teamid, teams[teamid].name);

	bprintf(buffer, 6, "www", CH_MSGFLAGDROP, p->id, teamid);
	DEBUG("::: BCAST_FLAGDROP %d %d", p->id, teamid);
	broadcast(buffer, 6);
}

void	bcast_flagreturn(struct player *p, int teamid)
{
	char	buffer[6];

	CONSOLE("%d::%s returns a flag of team %d::%s", p->id, p->name, teamid, teams[teamid].name);

	bprintf(buffer, 6, "www", CH_MSGFLAGRET, p->id, teamid);
	DEBUG("::: BCAST_FLAGRET %d %d", p->id, teamid);
	broadcast(buffer, 6);
}

void	bcast_flagwin(struct player *p, int teamid)
{
	char	buffer[6];

	CONSOLE("%d::%s delivers the flag of team %d::%s to his team's fortress!", p->id, p->name, teamid, teams[teamid].name);

	bprintf(buffer, 6, "www", CH_MSGFLAGWIN, p->id, teamid);
	DEBUG("::: BCAST_FLAGWIN %d %d", p->id, teamid);
	broadcast(buffer, 6);
}

/**********/

u16	newobj(void)
{
	return ++objcounter;
}

/**********/

void	parse(struct player *p)
{
	u16	opcode, datalen, port, stringlen;
	u8	*ptr;
	int	scanlen;
	char	buf[8];
	char	nickbuf[MAX_NICKLEN];
	u16    teamid;

	while(1) {
		struct buffer *pktbuf = p->tcp->in;

		ptr = pktbuf->data;
		datalen = BUFFER_NDATA(p->tcp->in);
		if (datalen < 2) {
			DEBUG("Too short packet %d<%d", datalen, 2);
			return;
		}
	
		opcode = UNPACKW(ptr);
		ptr += 2; datalen -= 2;
	
		DEBUG("[%d/%s] data: %d, opcode: %x", p->id, p->name, datalen, opcode);
		switch (opcode) {
			case CH_JOIN:
				stringlen = MAX_NICKLEN;
				BSCANF("join", "wwq", &p->fps, &teamid, &stringlen, &nickbuf);

				if (p->fps < 5 || p->fps > 25)
					p->fps = 25;

				p->name = ALLOC(stringlen + 1);
				strcpy(p->name, nickbuf);

				DEBUG("fps: %d port: %d name: %s", p->fps, port, p->name);
				join(p);
				join_team(p, teamid);
				bprintf(buf, 4, "ww", CH_MSGID, p->id);
				unicast(p, buf, 4);
				break;

			case CH_LEAVE:
				buffer_discard(pktbuf, 2);
				kick(p);
				break;

			case CH_PAUSE:
				buffer_discard(pktbuf, 2);
				DEBUG("Pause packet, cur. state %d", p->tank->status);
				if (p->tank->status != TANK_PAUSED)
					pause_tank(p->tank);
				else
					unpause_tank(p->tank);
				break;

			case CH_SAY:
			{
				char string[MAX_MSGLEN];

				stringlen = MAX_MSGLEN;
				BSCANF("say", "q", &stringlen, &string);

				bcast_say(p, string, stringlen);
			}
				break;

#if 0
			case CH_TEAMNAME:
				if (datalen < 6+msglen)
					return;
				teamid = UNPACKW(ptr); ptr+=2;
				if (teamid >= MAX_TEAMS)
					return;
				stringlen = msglen>=MAX_NICKLEN-1?MAX_NICKLEN-1:msglen;
				bcast_teamname(teamid, ptr, stringlen);
				buffer_discard(p->tcp->in, 6+msglen);
				break;
#endif

			case CH_TEAMUP:
				BSCANF("say", "w", &teamid);

				join_team(p, teamid);
				bcast_teamup(p, p->team);
				break;

			case CH_HELLO:
				break;
				
			default:
				DEBUG("UNKNOWN PACKET!");
				return;
		}
	}
}



/**********/

void	send_update(struct player *p)
{
	struct viewpoint *cvp, *lvp;
	u8	packet[2000], *pp;

	p->cur_seq++;

	if (p->last_seq + 10*FPS < p->cur_seq) {
		DEBUG("player %d is dead, kicking", p->id);
		kick(p);
		return;
	}
	
	cvp = addseq_q(&(p->state), p->cur_seq);
	lvp = loc_q(&(p->state), p->last_seq);

	if (! cvp || !lvp) {
		DEBUG("ERROR bad lvp or cvp");
		return;
	}
	tank_view(p->tank, cvp);
	pp = diff_view(lvp, cvp, packet, 2000);
	if (pp == NULL) {
		DEBUG("ERROR cannot diff states");
		return;
	}
	DEBUG("[%d] Sending update: c:%d, l:%d", p->id, p->cur_seq, p->last_seq);
	udppck_send_supdate(p->udp, p->cur_seq, p->last_seq, pp-packet, packet);
}

void	game_update(void)
{
	struct player *p, *n;
	tick++;

	DEBUG("game update!");
	move_game();
	foreachsafe(p, n, players)
		if (p->initialized && p->udp && !player_is_ai(p))
			send_update(p);
			
	tv.tv_sec = 0;
	tv.tv_usec = TICKMS*1000;
}

void	game_packet(struct player *p, struct udp_packet *pck)
{
	switch (pck->type) {
		case HC_UPDATE:
			DEBUG("*--> UPDATE %d: %d,(%d->%d)", p->id, pck->cu.direction, p->last_seq, pck->cu.ackseq);
			/* tady by se dalo nadelat bf overflowu */
			if (pck->cu.ackseq > p->cur_seq) {
				DEBUG("ERROR bad sequence number");
				return;
			}

			p->last_seq = pck->cu.ackseq;
			rmlt_q(&(p->state), p->last_seq);

			if (!tank_ready(p->tank))
				return;
			p->current_direction = pck->cu.direction;
			player_move(p, (p->current_direction)&7, (p->current_direction>>3)&15, (p->current_direction)&128);
			return;
	}
}

/**********/

void	mainloop(void)
{
	s32	ret;
	struct udp_packet 	packet;
	struct player	*p, *n;
	struct conn	*newc;
	int i;

	init_surface();
	init_game();
	init_teams();
	init_players();
	num_players = 0;
	for (i = 0; i < bots_to_spawn; i++)
		spawn_bot();

	tv.tv_sec = tv.tv_usec = 0;

	DEBUG("main loop");
	while (1) {
		fd_set 	fdr, fde;

		FD_ZERO(&fdr);
		FD_SET(server->fd, &fdr);
		FD_SET(chatserver->fd, &fdr);
		foreach (p, players)
			if (!player_is_ai(p))
				FD_SET(p->tcp->fd, &fdr);

		fde = fdr;
		ret = select(FD_SETSIZE, &fdr, NULL, &fde, &tv);
		if (ret < 0) {
			continue;
		} else if (ret == 0) {
			game_update();
		} else {
			 if (FD_ISSET(server->fd, &fdr)) {
				DEBUG("*** UDP PACKET");
			 	udppck_recv(server, &packet);
				foreachsafe(p, n, players) {
					if (!player_is_ai(p) && p->udp && udp_cmp(p->udp, &(packet.from)))
			 			game_packet(p, &packet);
				}
			} else if (FD_ISSET(chatserver->fd, &fdr)) {
				DEBUG("*** TCP NEW CONNECTION");
				newc = conn_accept(chatserver);
				conn_setbuff(newc, 900, 0);
				alloc_client(newc);
			} else {
				DEBUG("*** TCP ACTIVITY");
				foreachsafe (p, n, players) if (!player_is_ai(p) && FD_ISSET(p->tcp->fd, &fdr)) {
					int l;
					if ((l = buffer_read(p->tcp->in, p->tcp->fd)) <= 0) {
						DEBUG("`--- KICKING (%d) (%s)", l, l < 0 ?strerror(errno):"no data");
						kick(p);
					} else  {
						DEBUG("`--- PARSING data");
						parse(p);
					}
				}
			}
			foreachsafe (p, n, players) if (!player_is_ai(p) && FD_ISSET(p->tcp->fd, &fde)) {
				DEBUG("`--- KICKING, error");
				kick(p);
			}
		}
	}
}


static void
load_server_config(int *uport, int *tport)
{
	load_cfg_vars("server.cfg", "ai ferda udp-port tcp-port lms-lives botnames",
	                     "i i i i i as",
	                     &bots_to_spawn, &do_ferda, uport, tport,
	                     &lms_lives, &bot_names);
}

char version_banner[256];

static void
usage(char *argv[])
{
	printf("%s\n", version_banner);
	printf("Usage: %s [OPTION]... [TCPPORT]\n", argv[0]);
	printf("\n");
	printf("Mandatory arguments to long options are mandatory for short options too.\n");
	printf("\n");
	printf("Default TCPPORT value is %d.\n", DEFAULT_PORT);
	printf("\n");
	printf("General options:\n");
	printf("  -a, --ai=COUNT           spawn specified number of struct ai bots.\n");
	printf("  -f, --ferda              spawn Ferda the Zametac.\n");
	printf("  -P, --udp-port=UDPPORT   bind specified UDPPORT; by default the same\n");
	printf("                           port as TCPPORT is used, successive ports\n");
	printf("                           are probed on error.\n");
	printf("  -s, --game-style=STYLE   play the specified game style:\n");
	printf("                              dm    deathmatch (default)\n");
	printf("                              lms   last man standing\n");
	printf("                              ctf   capture the flag\n");
	printf("  -h, --help               show usage information.\n");
	printf("\n");
	printf("Teamplay-specific options:\n");
	printf("  -t, --teams-enforce[=NUMBER]   disallow freelance players; if such\n");
	printf("                                 a player connects, assign it to\n");
	printf("                                 a random one of the NUMBER teams\n");
	printf("                                 (default: %d if unspecified).\n", DEFAULT_FORCED_TEAMS);
	printf("      --teams-ai=NUMBER          spread struct ai to this number of teams evenly\n");
	printf("                                 (by default the AIs don't team up unless\n");
	printf("                                 -t was specified).\n");
	printf("\n");
	printf("Last man standing specific options:\n");
	printf("      --lms-lives=NUMBER   number of lives to final death (default: %d).\n", DEFAULT_LMS_LIVES);
	printf("\n");
	printf("Please report bugs and suggestions on #tunneler @ IRCNet.\n");
	exit(EX_USAGE);
}

int	main(int argc, char *argv[])
{
	int tport = -1, uport = -1;
	int show_usage = 0;

	signal(SIGPIPE, SIG_IGN);

	snprintf(version_banner, 256, "Tunneler server%s version udp-%04d -- in 2005 by ment, pasky, sd",
	         version_debug, version);

	load_server_config(&uport, &tport);

	while (1) {
#ifdef HAVE_GETOPT_LONG
		static struct option long_options[] = {
			{ "ai",         1, 0, 'a' },
			{ "ferda",      0, 0, 'f' },
			{ "help",       0, 0, 'h' },
			{ "udp-port",   1, 0, 'P' },
			{ "game-style", 1, 0, 's' },

			{ "teams-ai",   1, 0, 3 },
			{ "teams-enforce", 2, 0, 't' },

			{ "lms-lives",  1, 0, 2 },

			{ NULL,         0, 0, 0 }
		};
		int optindex;
#endif
		char *numguard = NULL;
		int c;
#define invalid_option(name_) \
		do { \
			fprintf(stderr, "%s: invalid "name_" -- %s\n", argv[0], optarg); \
			show_usage = 1; \
		} while (0)
#define load_number(name_, dest_) \
		do { \
			dest_ = strtol(optarg, &numguard, 10); \
			if (*numguard) invalid_option(name_); \
		} while (0)

#ifdef HAVE_GETOPT_LONG
		c = getopt_long(argc, argv, "-a:fhP:s:t::", long_options, &optindex);
#else
		c = getopt(argc, argv, "-a:fhP:s:t::");
#endif
		if (c == -1) break;

		switch (c) {
			case 1:   load_number("TCPPORT", tport); break;
			case 2:   load_number("NUMBER", lms_lives); break;
			case 3:   load_number("NUMBER", teams_ai);
				  if (teams_ai >= MAX_TEAMS) teams_ai = MAX_TEAMS - 1;
				  break;
			case 'a': load_number("COUNT", bots_to_spawn); break;
			case 'f': do_ferda = 1; break;
			case 'P': load_number("UDPPORT", uport); break;
			case 's':
				if (!strcmp(optarg, "dm")) {
					gstyle = GSTYLE_DM;
				} else if (!strcmp(optarg, "lms")) {
					gstyle = GSTYLE_LMS;
				} else if (!strcmp(optarg, "ctf")) {
					gstyle = GSTYLE_CTF;
				} else {
					invalid_option("STYLE");
				}
				break;
			case 't':
				if (!optarg) {
					teams_enforce = DEFAULT_FORCED_TEAMS;
				} else {
					load_number("NUMBER", teams_enforce);
					if (teams_enforce >= MAX_TEAMS)
						teams_enforce = MAX_TEAMS - 1;
				}
				break;
			case 'h':
			case '?': show_usage = 1; break;
		}
#undef load_number
#undef invalid_option
	}

	if (show_usage) usage(argv);
	if (tport < 0) tport = DEFAULT_PORT;
	if (uport < 0) uport = tport;


	debuglog("server.log");
	DEBUG("tunneler server starting...");
	CONSOLE("%s", version_banner);
	CONSOLE("Starting with resolution %dx%d, %d FPS...", VIEWPNT_WIDTH, VIEWPNT_HEIGHT, FPS);

	load_default_gfx();
	gfx_tank      = gfx_objects[GFX_TANK_ID];
	gfx_explosion = gfx_objects[GFX_EXPL_ID];
	gfx_bullet    = gfx_objects[GFX_BULLET_ID];
	gfx_home      = gfx_objects[GFX_HOME_ID];

	srandom(time(NULL));


	chatserver = conn_bind_port(tport, 5);
	if (chatserver->state != CS_LISTEN) {
		CONSOLE("Cannot bind port %d/tcp: %s", tport, strerror(errno));
		exit(EX_OSERR);
	}

	while (1) {
		server = udp_server(uport < 0 ? tport : uport);
		if (server->state == UCS_ERROR) {
			FREE(server);
			uport++;
		} else {
			break;
		}
	}
	
	CONSOLE("Bound ports %d/tcp and %d/udp.", tport, uport);

	mainloop();
	return 0;
}
